# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Transactions handling different types of requests."""

import abc
import collections
import copy
import sys
import time

from pinball.config.utils import get_log
from pinball.master.thrift_lib.ttypes import ArchiveRequest
from pinball.master.thrift_lib.ttypes import ErrorCode
from pinball.master.thrift_lib.ttypes import GroupRequest
from pinball.master.thrift_lib.ttypes import GroupResponse
from pinball.master.thrift_lib.ttypes import ModifyRequest
from pinball.master.thrift_lib.ttypes import ModifyResponse
from pinball.master.thrift_lib.ttypes import QueryAndOwnRequest
from pinball.master.thrift_lib.ttypes import QueryAndOwnResponse
from pinball.master.thrift_lib.ttypes import QueryRequest
from pinball.master.thrift_lib.ttypes import QueryResponse
from pinball.master.thrift_lib.ttypes import TokenMasterException


__author__ = 'Pawel Garbacki'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


LOG = get_log('pinball.master.transaction')


class Transaction(object):
    """Interface defining a transaction on a token trie."""
    __metaclass__ = abc.ABCMeta

    def __init__(self):
        self._updates = []
        self._deletes = []
        self._committed = False
        self._blessed_version = None
        self._store = None
        self._trie = None

    @abc.abstractmethod
    def prepare(self, request):
        """Prepare to handle a provided request."""
        return

    @abc.abstractmethod
    def commit(self, trie, blessed_version, store):
        """Execute the transaction on the provided token trie."""
        return

    @staticmethod
    def _verify_have_version(tokens):
        """Raise an exception if any token does not have a version set."""
        for token in tokens:
            if not token.version:
                raise TokenMasterException(
                    ErrorCode.INPUT_ERROR,
                    'token %s does not have version set' % token.name)

    def _set_trie(self, trie, blessed_version, store):
        self._trie = trie
        self._blessed_version = copy.copy(blessed_version)
        self._store = store

    def _add_update(self, token):
        """Add a token update resetting the token version."""
        token = copy.copy(token)
        token.version = self._blessed_version.advance_version()
        self._updates.append(token)

    def _add_delete(self, token):
        self._deletes.append(token)

    def _commit(self):
        """Merge token updates into the trie."""
        assert not self._committed
        try:
            self._blessed_version.advance_version()
            self._trie[self._blessed_version.name] = self._blessed_version
            self._store.commit_tokens(self._updates + [self._blessed_version],
                                      self._deletes)
            for token in self._updates:
                self._trie[token.name] = token
            for token in self._deletes:
                del self._trie[token.name]
        except:
            # This should never happen but if it does happen, our state will
            # get out of sync so we better crash.
            LOG.exception('')
            sys.exit(1)
        self._committed = True


class ArchiveTransaction(Transaction):
    """Transaction handling archive requests."""
    def __init__(self):
        super(ArchiveTransaction, self).__init__()
        self._request = None

    def prepare(self, request):
        self._request = request
        if request.tokens:
            Transaction._verify_have_version(request.tokens)

    def _verify_archive_tokens(self):
        for token in self._request.tokens:
            existing_token = self._trie.get(token.name)
            if not existing_token:
                raise TokenMasterException(ErrorCode.NOT_FOUND,
                                           'token %s not found' % token.name)
            if token.version != existing_token.version:
                raise TokenMasterException(
                    ErrorCode.VERSION_CONFLICT,
                    'token %s with different version %d found' %
                    (existing_token.name, existing_token.version))

    def commit(self, trie, blessed_version, store):
        self._set_trie(trie, blessed_version, store)
        if self._request.tokens:
            self._verify_archive_tokens()
            try:
                store.archive_tokens(self._request.tokens)
                for token in self._request.tokens:
                    del self._trie[token.name]
            except:
                # This should never happen but if it does happen, our state
                # will get out of sync so we better crash.
                LOG.exception('')
                sys.exit(1)


class GroupTransaction(Transaction):
    """Transaction handling group requests."""
    def __init__(self):
        super(GroupTransaction, self).__init__()
        self._request = None

    def prepare(self, request):
        self._request = request

    def _get_group(self, name):
        assert name.startswith(self._request.namePrefix)
        start = len(self._request.namePrefix)
        if self._request.groupSuffix:
            end = name.find(self._request.groupSuffix, start)
        else:
            end = len(name)
        if end == -1:
            end = len(name)
        return name[:end]

    def commit(self, trie, blessed_version, store):
        response = GroupResponse()
        if self._request.namePrefix:
            response.counts = collections.defaultdict(int)
            names = trie.keys(self._request.namePrefix)
            for name in names:
                group = self._get_group(name)
                response.counts[group] += 1
        return response


class ModifyTransaction(Transaction):
    """Transaction handling update requests."""
    def __init__(self):
        super(ModifyTransaction, self).__init__()
        self._request = None

    def prepare(self, request):
        self._request = request
        if self._request.deletes:
            Transaction._verify_have_version(self._request.deletes)
        # TODO(pawel): add more checks of the request

    def _verify_modify_tokens(self, tokens):
        """Check if tokens on the provided list can be modified."""
        for token in tokens:
            if token.name in self._trie:
                existing_token = self._trie[token.name]
                assert token.name == existing_token.name
                assert existing_token.version
                if existing_token.version != token.version:
                    raise TokenMasterException(
                        ErrorCode.VERSION_CONFLICT,
                        'token %s with different version %d found' %
                        (existing_token.name, existing_token.version))
            elif token.version:
                raise TokenMasterException(ErrorCode.NOT_FOUND,
                                           'token %s not found' % token.name)

    def commit(self, trie, blessed_version, store):
        self._set_trie(trie, blessed_version, store)
        if self._request.updates:
            self._verify_modify_tokens(self._request.updates)
        if self._request.deletes:
            self._verify_modify_tokens(self._request.deletes)
        if self._request.updates:
            for token in self._request.updates:
                self._add_update(token)
        if self._request.deletes:
            for token in self._request.deletes:
                self._add_delete(token)
        self._commit()
        response = ModifyResponse()
        if self._updates:
            response.updates = []
        for token in self._updates:
            response.updates.append(token)
        return response


class QueryTransaction(Transaction):
    """Transaction handling query requests."""
    def __init__(self):
        super(QueryTransaction, self).__init__()
        self._request = None

    def prepare(self, request):
        self._request = request
        # TODO(pawel): add checks of the request

    @staticmethod
    def _sort_on_priority(tokens):
        def _priority(token):
            if not token.priority:
                return 0
            return token.priority
        return sorted(tokens, key=_priority, reverse=True)

    def _get_tokens(self, query):
        """Retrieve tokens matching a given query."""
        matching_tokens = self._trie.values(query.namePrefix)
        if query.maxTokens is None:
            return matching_tokens
        sorted_tokens = QueryTransaction._sort_on_priority(matching_tokens)
        return sorted_tokens[:query.maxTokens]

    def commit(self, trie, blessed_version, store):
        self._set_trie(trie, blessed_version, store)
        response = QueryResponse()
        if self._request.queries:
            response.tokens = []
            for query in self._request.queries:
                response.tokens.append(self._get_tokens(query))
        return response


class QueryAndOwnTransaction(QueryTransaction):
    """Transaction handling query and own requests."""
    @staticmethod
    def _get_timestamp_secs():
        """Return time in seconds since the epoch."""
        return int(time.time())

    @ staticmethod
    def _is_owned(token):
        """Check if a given token is currently owned."""
        return (token.owner and
                token.expirationTime and
                (token.expirationTime >
                 QueryAndOwnTransaction._get_timestamp_secs()))

    def commit(self, trie, blessed_version, store):
        self._set_trie(trie, blessed_version, store)
        response = QueryAndOwnResponse()
        response.tokens = []
        if self._request.query:
            matching_tokens = self._trie.values(self._request.query.namePrefix)
            sorted_tokens = QueryTransaction._sort_on_priority(matching_tokens)
            for token in sorted_tokens:
                if (self._request.query.maxTokens is not None and
                        len(self._updates) >= self._request.query.maxTokens):
                    break
                if not QueryAndOwnTransaction._is_owned(token):
                    token = copy.copy(token)
                    token.owner = self._request.owner
                    token.expirationTime = self._request.expirationTime
                    self._add_update(token)
            self._commit()
            response.tokens = self._updates
        return response


# Mapping from request class to the transaction class that handles requests of
# this type.
REQUEST_TO_TRANSACTION = {ArchiveRequest: ArchiveTransaction,
                          GroupRequest: GroupTransaction,
                          ModifyRequest: ModifyTransaction,
                          QueryAndOwnRequest: QueryAndOwnTransaction,
                          QueryRequest: QueryTransaction}



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Command line tool to monitor and control workflows."""
import abc
import argparse
import datetime
import getpass
import pickle
import shutil
import socket
import sys
import time

from pinball.common.s3_utils import delete_s3_directory
from pinball.config.pinball_config import PinballConfig
from pinball.config.utils import get_unique_name
from pinball.config.utils import master_name
from pinball.master.factory import Factory
from pinball.master.thrift_lib.ttypes import GroupRequest
from pinball.master.thrift_lib.ttypes import ModifyRequest
from pinball.master.thrift_lib.ttypes import Query
from pinball.master.thrift_lib.ttypes import QueryAndOwnRequest
from pinball.master.thrift_lib.ttypes import QueryRequest
from pinball.master.thrift_lib.ttypes import Token
from pinball.parser.config_parser import ParserCaller
from pinball.parser.utils import load_parser_with_caller
from pinball.tools.base import Command
from pinball.tools.base import CommandException
from pinball.tools.base import confirm
from pinball.ui.data import Status
from pinball.ui.data_builder import DataBuilder
from pinball.workflow.analyzer import Analyzer
from pinball.workflow.name import Name
from pinball.workflow.signaller import Signal
from pinball.workflow.signaller import Signaller
from pinball.workflow.utils import get_logs_dir
from pinball.workflow.utils import get_unique_workflow_instance


__author__ = 'Pawel Garbacki, Mao Ye'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


def _get_all_tokens(workflow, instance, client):
    """Get all tokens in a given workflow instance.

    Args:
        workflow: The name of the workflow whose tokens should be returned.
        instance: The workflow instance whose tokens should be returned.
        client: The client to use when communicating with the master.
    Returns:
        List of tokens in a given workflow instance.
    """
    name = Name(workflow=workflow, instance=instance)
    prefix = name.get_instance_prefix()
    query = Query(namePrefix=prefix)
    request = QueryRequest(queries=[query])
    response = client.query(request)
    assert len(response.tokens) == 1
    return response.tokens[0]


class Start(Command):
    """Parse and start a new workflow instance."""
    def __init__(self):
        self._workflow = None

    def prepare(self, options):
        self._workflow = options.workflow
        if not self._workflow:
            raise CommandException('start command takes name of '
                                   'workflow to start')

    def execute(self, client, store):
        if not _check_workflow_instances(store, self._workflow):
            raise CommandException('failed to start a new instance for '
                                   'workflow %s due to too many instances '
                                   'running!' % self._workflow)

        config_parser = load_parser_with_caller(PinballConfig.PARSER,
                                                PinballConfig.PARSER_PARAMS,
                                                ParserCaller.WORKFLOW_UTIL)
        workflow_tokens = config_parser.get_workflow_tokens(self._workflow)
        if not workflow_tokens:
            return 'workflow %s not found in %s\n' % (
                self._workflow, str(PinballConfig.PARSER_PARAMS))
        request = ModifyRequest()
        request.updates = workflow_tokens
        assert request.updates
        token = request.updates[0]
        name = Name.from_job_token_name(token.name)
        if not name.instance:
            name = Name.from_event_token_name(token.name)
        client.modify(request)
        return 'exported workflow %s instance %s.  Its tokens are under %s' % (
            name.workflow, name.instance, name.get_instance_prefix())


class Stop(Command):
    """Stop a running workflow instance."""
    def __init__(self):
        self._workflow = None
        self._force = None
        self._instance = None

    def prepare(self, options):
        self._force = options.force
        self._workflow = options.workflow
        self._instance = options.instance
        if not self._workflow or not self._instance:
            raise CommandException('stop command takes name of '
                                   'workflow and instance')

    def execute(self, client, store):
        instance_tokens = _get_all_tokens(self._workflow,
                                          self._instance,
                                          client)
        if not instance_tokens:
            return 'workflow %s instance %s not found\n' % (self._workflow,
                                                            self._instance)
        message = 'Remove workflow %s instance %s' % (self._workflow,
                                                      self._instance)
        output = ''
        if self._force or confirm(message):
            MAX_TRIES = 10
            i = 0
            while i < MAX_TRIES:
                i += 1
                request = ModifyRequest(deletes=instance_tokens)
                try:
                    client.modify(request)
                    break
                except:
                    # This can happen because someone else could have updated
                    # one of the tokens in the meantime.
                    # TODO(pawel): come up with a mechanism to prevent the need
                    # to retry.
                    instance_tokens = _get_all_tokens(self._workflow,
                                                      self._instance,
                                                      client)
            output += 'removed %d token(s) in %d tries\n' % (
                len(instance_tokens), i)
        return output


class Pause(Command):
    """Pause a running workflow instance."""
    def __init__(self):
        self._workflow = None
        self._force = None
        self._instance = None

    def prepare(self, options):
        self._force = options.force
        self._workflow = options.workflow
        self._instance = options.instance
        if not self._workflow or not self._instance:
            raise CommandException('pause command takes name of '
                                   'workflow and instance')

    @staticmethod
    def _own_tokens(tokens):
        for token in tokens:
            timestamp = datetime.datetime.now().strftime("%Y-%m-%d %H:%M")
            token.owner = 'workflow_util user %s machine %s time %s' % (
                getpass.getuser(), socket.gethostname(), timestamp)
            token.expirationTime = sys.maxint

    def execute(self, client, store):
        instance_tokens = _get_all_tokens(self._workflow,
                                          self._instance,
                                          client)
        if not instance_tokens:
            return 'workflow %s instance %s not found\n' % (self._workflow,
                                                            self._instance)
        message = 'pause workflow %s instance %s' % (self._workflow,
                                                     self._instance)
        output = ''
        if self._force or confirm(message):
            MAX_TRIES = 10
            i = 0
            while i < MAX_TRIES:
                Pause._own_tokens(instance_tokens)
                request = ModifyRequest(updates=instance_tokens)
                try:
                    client.modify(request)
                    break
                except:
                    # This can happen because someone else could have updated
                    # one of the tokens in the meantime.
                    # TODO(pawel): come up with a mechanism to prevent the need
                    # to retry.
                    instance_tokens = _get_all_tokens(self._workflow,
                                                      self._instance,
                                                      client)
                i += 1
            if i < MAX_TRIES:
                output += 'claimed %d token(s) in %d tries\n' % (
                    len(instance_tokens), i + 1)
            else:
                output += 'failed to claim token(s) in %d tries' % MAX_TRIES
        return output


class Resume(Command):
    """Resume a paused workflow instance."""
    def __init__(self):
        self._workflow = None
        self._force = None
        self._instance = None

    def prepare(self, options):
        self._force = options.force
        self._workflow = options.workflow
        self._instance = options.instance
        if not self._workflow or not self._instance:
            raise CommandException('resume command takes name of '
                                   'workflow and instance')

    def execute(self, client, store):
        instance_tokens = _get_all_tokens(self._workflow,
                                          self._instance,
                                          client)
        if not instance_tokens:
            return 'workflow %s instance %s not found\n' % (self._workflow,
                                                            self._instance)
        message = 'resume workflow %s instance %s' % (self._workflow,
                                                      self._instance)
        output = ''
        if self._force or confirm(message):
            for token in instance_tokens:
                token.owner = None
                token.expirationTime = None
            request = ModifyRequest(updates=instance_tokens)
            client.modify(request)
            output += ('released ownership of %d token(s)\n' %
                       len(instance_tokens))
        return output


class Retry(Command):
    """Retry failed jobs."""
    def __init__(self):
        self._workflow = None
        self._force = None
        self._instance = None

    def prepare(self, options):
        self._force = options.force
        self._workflow = options.workflow
        self._instance = options.instance
        if not self._workflow or not self._instance:
            raise CommandException('retry command takes name of workflow and '
                                   'instance')

    @staticmethod
    def _is_job_failed(job_token):
        job = pickle.loads(job_token.data)
        if not job.history:
            return False
        last_execution_record = job.history[-1]
        return last_execution_record.exit_code != 0

    @staticmethod
    def _prepare_runnable_job(job):
        if job.history:
            last_execution_record = job.history[-1]
            if last_execution_record.events:
                job.events = last_execution_record.events

    def _retry_active(self, client):
        # Retrieve waiting jobs and ARCHIVE signal from the master.
        query_request = QueryRequest(queries=[])

        signal_name = Name(workflow=self._workflow, instance=self._instance,
                           signal=Signal.action_to_string(Signal.ARCHIVE))
        signal_prefix = signal_name.get_signal_token_name()
        query_request.queries.append(Query(namePrefix=signal_prefix))

        jobs_name = Name(workflow=self._workflow, instance=self._instance,
                         job_state=Name.WAITING_STATE)
        jobs_prefix = jobs_name.get_job_state_prefix()
        query_request.queries.append(Query(namePrefix=jobs_prefix))

        query_response = client.query(query_request)
        assert len(query_response.tokens) == 2

        modify_request = ModifyRequest(updates=[], deletes=[])

        # Remove the ARCHIVE signal.
        if query_response.tokens[0]:
            assert len(query_response.tokens[0]) == 1
            signal_token = query_response.tokens[0][0]
            modify_request.deletes.append(signal_token)

        # Make failed jobs runnable.
        for job_token in query_response.tokens[1]:
            if Retry._is_job_failed(job_token):
                modify_request.deletes.append(job_token)
                runnable_job_name = Name.from_job_token_name(job_token.name)
                runnable_job_name.job_state = Name.RUNNABLE_STATE
                runnable_job = pickle.loads(job_token.data)
                self._prepare_runnable_job(runnable_job)
                runnable_job_token = Token(
                    name=runnable_job_name.get_job_token_name(),
                    priority=job_token.priority,
                    data=pickle.dumps(runnable_job))
                modify_request.updates.append(runnable_job_token)

        if not modify_request.updates and not modify_request.deletes:
            return 'no failed jobs found in workflow %s instance %s\n' % (
                self._workflow, self._instance)
        if not modify_request.updates and modify_request.deletes:
            return 'found ARCHIVE token but no failed jobs in workflow %s ' \
                   'instance %s.  Not changing anything this time\n' % (
                       self._workflow, self._instance)

        client.modify(modify_request)

        if len(modify_request.updates) == len(modify_request.deletes):
            return 'retried %d job(s) and removed an ARCHIVE token from ' \
                   'workflow %s instance %s\n' % (len(modify_request),
                                                  self._workflow,
                                                  self._instance)
        assert (len(modify_request.updates) == len(modify_request.deletes) - 1)
        return 'retried %d job(s) in workflow %s instance %s\n' % (
            len(modify_request.updates), self._workflow, self._instance)

    def _retry_archived(self, client, store):
        instance_name = Name(workflow=self._workflow, instance=self._instance)
        instance_tokens = store.read_archived_tokens(
            name_prefix=instance_name.get_instance_prefix())
        if not instance_tokens:
            return 'workflow %s instance %s not found\n' % (self._workflow,
                                                            self._instance)

        request = ModifyRequest(updates=[])
        new_instance = get_unique_workflow_instance()

        has_failed_jobs = False
        for token in instance_tokens:
            token_name = token.name

            event_name = Name.from_event_token_name(token_name)
            if event_name.workflow:
                # it is an event token
                event_name.instance = new_instance
                event_token = Token(name=event_name.get_event_token_name(),
                                    priority=token.priority,
                                    data=token.data)
                request.updates.append(event_token)
                continue

            job_name = Name.from_job_token_name(token_name)
            if job_name.workflow:
                # it is a job token
                if Retry._is_job_failed(token):
                    has_failed_jobs = True
                    job_name.job_state = Name.RUNNABLE_STATE
                job_name.instance = new_instance
                job_token = Token(name=job_name.get_job_token_name(),
                                  priority=token.priority,
                                  data=token.data)
                request.updates.append(job_token)
                continue

            signal_name = Name.from_signal_token_name(token_name)
            assert signal_name.workflow
            # it is a signal token.  We ignore those.

        if not has_failed_jobs:
            return 'no failed jobs found in workflow %s instance %s\n' % (
                self._workflow, self._instance)

        client.modify(request)
        new_instance_name = Name(workflow=self._workflow,
                                 instance=new_instance)
        return 'retried workflow %s instance %s.  Its tokens are under ' \
               '%s\n' % (self._workflow,
                         self._instance,
                         new_instance_name.get_instance_prefix())

    def execute(self, client, store):
        if not _check_workflow_instances(store, self._workflow):
            raise CommandException('failed to retry workflow %s instance %s due '
                                   'to too many instances are running!' %
                                   (self._workflow, self._instance))

        message = 'retry workflow %s instance %s' % (self._workflow,
                                                     self._instance)
        output = ''
        if self._force or confirm(message):
            name = Name(workflow=self._workflow, instance=self._instance)
            request = GroupRequest(namePrefix=name.get_job_prefix(),
                                   groupSuffix=Name.DELIMITER)
            response = client.group(request)
            if response.counts:
                output += self._retry_active(client)
            else:
                output += self._retry_archived(client, store)
        return output


class Redo(Command):
    """Redo a previously executed job."""
    def __init__(self):
        self._force = None
        self._workflow = None
        self._instance = None
        self._job = None
        self._execution = None

    def prepare(self, options):
        self._force = options.force
        self._workflow = options.workflow
        self._instance = options.instance
        self._job = options.jobs
        self._execution = int(options.execution)
        if ',' in self._job or ' ' in self._job:
            raise CommandException('redo command takes a single job')
        if (not self._workflow or not self._instance or not self._job):
            raise CommandException('redo command takes name of workflow, '
                                   'instance, job, and execution')

    def _get_execution_record(self, job):
        if not job.history:
            return None
        if len(job.history) <= self._execution:
            return None
        return job.history[self._execution]

    def execute(self, client, store):
        output = ''
        message = 'redo execution %d of job %s in workflow %s instance %s' % (
            self._execution, self._job, self._workflow, self._instance)
        if self._force or confirm(message):
            # Retrieve job token.
            owner = get_unique_name()
            expiration_time = time.time() + 60
            job_name = Name(workflow=self._workflow, instance=self._instance,
                            job_state=Name.WAITING_STATE, job=self._job)
            query = Query(namePrefix=job_name.get_job_token_name())
            query_and_own_request = QueryAndOwnRequest(
                owner=owner, expirationTime=expiration_time, query=query)
            query_and_own_response = client.query_and_own(
                query_and_own_request)
            if not query_and_own_response.tokens:
                return 'workflow must be running, the job must be finished ' \
                       'and it cannot be runnable'

            # Find the job token.
            modify_request = ModifyRequest(updates=[])
            waiting_job = None
            for job_token in query_and_own_response.tokens:
                token_name = Name.from_job_token_name(job_token.name)
                if token_name.job == self._job:
                    assert not waiting_job
                    waiting_job = job_token
                else:
                    job_token.owner = None
                    job_token.expirationTime = None
                    modify_request.updates.append(job_token)

            # Make the job runnable.
            job = pickle.loads(waiting_job.data)
            execution_record = self._get_execution_record(job)
            if not execution_record:
                # Unown the job token.
                waiting_job.owner = None
                waiting_job.expirationTime = None
                modify_request.updates.append(waiting_job)
                output = ('could not find execution %d in job history\n' %
                          self._execution)
            else:
                job_name.job_state = Name.RUNNABLE_STATE
                job.events = execution_record.events
                runnable_job = Token(name=job_name.get_job_token_name(),
                                     data=pickle.dumps(job))
                modify_request.updates.append(runnable_job)
                modify_request.deletes = [waiting_job]
                output = ('redoing execution %d of job %s in workflow %s '
                          'instance %s\n' % (self._execution, self._job,
                                             self._workflow, self._instance))
            client.modify(modify_request)

        return output


class Poison(Command):
    """Run poisoned jobs."""
    def __init__(self):
        self._workflow = None
        self._force = None
        self._jobs = None
        self._instance = None

    def prepare(self, options):
        self._force = options.force
        self._workflow = options.workflow
        self._instance = options.instance
        self._jobs = options.jobs.split(',')
        if (not (self._workflow and self._instance and self._jobs) and
                not (self._workflow and self._jobs and
                     PinballConfig.PARSER_PARAMS)):
            raise CommandException('retry command takes name of workflow, '
                                   'instance, and a list of jobs or workflow, '
                                   'parser_param config, and a list of jobs')

    def _get_archive_token(self, client):
        assert self._instance
        name = Name(workflow=self._workflow,
                    instance=self._instance,
                    signal=Signal.action_to_string(Signal.ARCHIVE))
        prefix = name.get_signal_token_name()
        query = Query(namePrefix=prefix)
        request = QueryRequest(queries=[query])
        response = client.query(request)
        assert len(response.tokens) == 1
        if not response.tokens[0]:
            return None
        assert len(response.tokens[0]) == 1
        return response.tokens[0][0]

    def _poison_active(self, client):
        analyzer = Analyzer.from_client(client, self._workflow, self._instance)
        if not analyzer.get_tokens():
            return 'workflow %s instance %s not found\n' % (self._workflow,
                                                            self._instance)
        analyzer.poison(self._jobs)
        event_tokens = analyzer.get_new_event_tokens()
        archive_token = self._get_archive_token(client)
        archive_tokens = []
        if archive_token:
            archive_tokens.append(archive_token)
        request = ModifyRequest(updates=event_tokens, deletes=archive_tokens)

        client.modify(request)
        return 'poisoned workflow %s instance %s roots %s\n' % (
            self._workflow, self._instance, self._jobs)

    def _poison_inactive(self, client, store):
        if self._instance:
            analyzer = Analyzer.from_store(store,
                                           self._workflow,
                                           self._instance)
            if not analyzer.get_tokens():
                return 'workflow %s instance %s not found\n' % (
                    self._workflow, self._instance)
        else:
            analyzer = Analyzer.from_parser_params(self._workflow)
            if not analyzer.get_tokens():
                return 'workflow %s not found in %s\n' % (
                    self._workflow, str(PinballConfig.PARSER_PARAMS))
        analyzer.clear_job_histories()
        analyzer.poison(self._jobs)
        new_instance = get_unique_workflow_instance()
        analyzer.change_instance(new_instance)
        tokens = analyzer.get_tokens()
        request = ModifyRequest(updates=tokens)
        client.modify(request)
        new_instance_name = Name(workflow=self._workflow,
                                 instance=new_instance)
        return 'poisoned workflow %s roots %s.  Tokens of the new ' \
               'instance are under %s\n' % (
                   self._workflow,
                   self._jobs,
                   new_instance_name.get_instance_prefix())

    def execute(self, client, store):
        if self._instance:
            message = 'poison workflow %s instance %s roots %s' % (
                self._workflow, self._instance, self._jobs)
        else:
            message = 'poison workflow %s roots %s parser_params config %s' % (
                self._workflow, self._jobs, str(PinballConfig.PARSER_PARAMS))
        if self._force or confirm(message):
            active = False
            if self._instance:
                name = Name(workflow=self._workflow, instance=self._instance)
                request = GroupRequest(namePrefix=name.get_job_prefix(),
                                       groupSuffix=Name.DELIMITER)
                response = client.group(request)
                if response.counts:
                    active = True
            if active:
                return self._poison_active(client)
            else:
                return self._poison_inactive(client, store)
        return ''


class ModifySignal(Command):
    """Add or remove a signal.

    A signal is added to or removed from the master.
    """
    __metaclass__ = abc.ABCMeta

    def __init__(self, action, add):
        self._action = action
        # If not add then remove.
        self._add = add
        self._workflow = None
        self._force = None
        self._instance = None

    def prepare(self, options):
        self._force = options.force
        self._workflow = options.workflow
        self._instance = options.instance
        if not self._workflow and self._instance:
            raise CommandException('workflow must be provided if instance is '
                                   'set')

    def execute(self, client, store):
        action_name = Signal.action_to_string(self._action)
        if self._add:
            # Prefix to put in front of the action name to differentiate
            # between adding and removing the action.
            action_prefix = ''
        else:
            action_prefix = 'UN'
        if self._instance:
            assert self._workflow
            message = '%s%s workflow %s instance %s' % (action_prefix,
                                                        action_name,
                                                        self._workflow,
                                                        self._instance)
        elif self._workflow:
            message = '%s%s all instances of workflow %s' % (action_prefix,
                                                             action_name,
                                                             self._workflow)
        else:
            message = '%s%s all workflows' % (action_prefix, action_name)
        output = ''
        if self._force or confirm(message):
            if self._workflow:
                workflow = self._workflow
            else:
                workflow = None
            if self._instance:
                instance = self._instance
            else:
                instance = None
            signaller = Signaller(client, workflow, instance)
            if self._add and signaller.is_signal_present(self._action):
                output += ('%s has been already set.  Not changing '
                           'anything this time\n' % action_name)
            elif (not self._add and
                  not signaller.is_signal_present(self._action)):
                output += ('%s has been already removed.  Not changing '
                           'anything this time\n' % action_name)
            else:
                name = Name(workflow=workflow, instance=instance,
                            signal=action_name)
                if self._add:
                    signaller.set_action(self._action)
                    output += ('set %s.  Its token is %s\n' % (action_name,
                               name.get_signal_token_name()))
                else:
                    signaller.remove_action(self._action)
                    output += ('removed %s from %s\n' % (action_name,
                               name.get_signal_token_name()))
        return output


class Drain(ModifySignal):
    """Stop kicking off new jobs."""
    def __init__(self):
        super(Drain, self).__init__(Signal.DRAIN, True)


class UnDrain(ModifySignal):
    """Start accepting new jobs after calling Drain."""
    def __init__(self):
        super(UnDrain, self).__init__(Signal.DRAIN, False)


class Abort(ModifySignal):
    """Cancel all jobs running jobs, fail workflows."""
    def __init__(self):
        super(Abort, self).__init__(Signal.ABORT, True)


class UnAbort(ModifySignal):
    """Stop canceling jobs and failing workflows after calling Abort."""
    def __init__(self):
        super(UnAbort, self).__init__(Signal.ABORT, False)


class Exit(ModifySignal):
    """Exit the workers without failing workflows."""
    def __init__(self):
        super(Exit, self).__init__(Signal.EXIT, True)


class UnExit(ModifySignal):
    """Stop exiting workers after calling Exit."""
    def __init__(self):
        super(UnExit, self).__init__(Signal.EXIT, False)


class ModifySchedule(Command):
    """Base class for operations on workflow schedules."""
    __metaclass__ = abc.ABCMeta

    def __init__(self):
        self._workflow = None

    @staticmethod
    def _get_schedule_tokens(client, workflows):
        request = QueryRequest(queries=[])
        for workflow in workflows:
            name = Name(workflow=workflow)
            prefix = name.get_workflow_schedule_token_name()
            query = Query(namePrefix=prefix)
            request.queries.append(query)
        response = client.query(request)
        assert len(response.tokens) == len(workflows)
        result = []
        for i in range(0, len(workflows)):
            schedule_tokens = response.tokens[i]
            if schedule_tokens:
                token_found = False
                for schedule_token in schedule_tokens:
                    name = Name.from_workflow_schedule_token_name(
                        schedule_token.name)
                    if name.workflow == workflows[i]:
                        result.append(schedule_token)
                        token_found = True
                        break
                if not token_found:
                    result.append(None)
            else:
                result.append(None)
        return result


class ReSchedule(ModifySchedule):
    """Update workflow schedule.

    Replace the workflow schedule stored in the master with one parsed from
    the configuration.
    """
    def __init__(self):
        super(ReSchedule, self).__init__()
        self._force = None

    def prepare(self, options):
        self._workflow = options.workflow
        self._force = options.force

    def execute(self, client, store):
        config_parser = load_parser_with_caller(PinballConfig.PARSER,
                                                PinballConfig.PARSER_PARAMS,
                                                ParserCaller.WORKFLOW_UTIL)
        workflow_names = config_parser.get_workflow_names()
        if (self._workflow and not self._workflow in workflow_names):
            return 'workflow %s not found\n' % self._workflow
        workflows = ([self._workflow] if self._workflow else workflow_names)
        if not workflows:
            return 'no workflows found in %s' % str(PinballConfig.PARSER_PARAMS)
        message = 'reschedule workflows %s' % workflows
        output = ''
        if self._force or confirm(message):
            request = ModifyRequest(updates=[])
            old_schedule_tokens = ModifySchedule._get_schedule_tokens(
                client, workflows)
            schedule_token_names = []
            rescheduled_workflows = []
            for i in range(0, len(workflows)):
                workflow = workflows[i]
                new_schedule_token = config_parser.get_schedule_token(workflow)
                if not new_schedule_token:
                    continue

                old_schedule_token = old_schedule_tokens[i]
                old_corresponds_new = False
                if old_schedule_token:
                    assert old_schedule_token.name == new_schedule_token.name
                    new_schedule_token.version = old_schedule_token.version
                    old_schedule = pickle.loads(old_schedule_token.data)
                    new_schedule = pickle.loads(new_schedule_token.data)
                    old_corresponds_new = old_schedule.corresponds_to(new_schedule)

                if not old_schedule_token or not old_corresponds_new:
                    request.updates.append(new_schedule_token)
                    schedule_token_names.append(new_schedule_token.name)
                    rescheduled_workflows.append(workflow)
            client.modify(request)
            output = ('rescheduled workflows %s.  Their schedule tokens are '
                      '%s\n' % (rescheduled_workflows, schedule_token_names))
        return output


class UnSchedule(ModifySchedule):
    """Remove a workflow from the schedule."""
    def __init__(self):
        super(UnSchedule, self).__init__()
        self._force = None

    def prepare(self, options):
        self._workflow = options.workflow
        if not self._workflow:
            raise CommandException('unschedule command takes name of workflow '
                                   'to remove from the schedule')
        self._force = options.force

    def execute(self, client, store):
        schedule_tokens = self._get_schedule_tokens(client, [self._workflow])
        assert len(schedule_tokens) == 1
        schedule_token = schedule_tokens[0]
        if not schedule_token:
            return 'schedule for workflow %s not found\n' % self._workflow
        message = 'remove schedule for workflow %s' % self._workflow
        output = ''
        if self._force or confirm(message):
            request = ModifyRequest(deletes=[schedule_token])
            client.modify(request)
            output = 'removed schedule for workflow %s\n' % self._workflow
        return output


class Reload(Command):
    """Reload jobs configuration."""
    _LEASE_TIME_SEC = 5 * 60

    def __init__(self):
        super(Reload, self).__init__()
        self._workflow = None
        self._instance = None
        self._jobs = None
        self._force = None
        self._output = ''

    def prepare(self, options):
        self._force = options.force
        self._workflow = options.workflow
        self._instance = options.instance
        self._jobs = options.jobs.split(',') if options.jobs else None
        if (not self._workflow or not self._instance or
                not PinballConfig.PARSER_PARAMS):
            raise CommandException('reload command requires workflow name, '
                                   'instance, and parser_params config')

    @staticmethod
    def _unown_tokens(client, tokens):
        if not tokens:
            return
        for token in tokens:
            token.owner = None
            token.expirationTime = None
        request = ModifyRequest(updates=tokens)
        client.modify(request)

    @staticmethod
    def _reload_job_token(job_token, new_job_token):
        job_token.priority = new_job_token.priority
        job = pickle.loads(job_token.data)
        new_job = pickle.loads(new_job_token.data)
        job.reload(new_job)
        job_token.data = pickle.dumps(job)

    def _own_selected_job_tokens(self, client):
        assert self._jobs
        result = []
        for job in self._jobs:
            token = None
            for state in [Name.RUNNABLE_STATE, Name.WAITING_STATE]:
                name = Name(workflow=self._workflow, instance=self._instance,
                            job_state=state, job=job)
                query = Query(namePrefix=name.get_job_token_name(),
                              maxTokens=1)
                request = QueryAndOwnRequest(
                    owner='workflow_util',
                    expirationTime=time.time() + Reload._LEASE_TIME_SEC,
                    query=query)
                response = client.query_and_own(request)
                if response.tokens:
                    assert len(response.tokens) == 1
                    if response.tokens[0].name == name.get_job_token_name():
                        token = response.tokens[0]
                        break
            if token:
                result.append(token)
            else:
                self._output += ('job %s in workflow %s instance %s either '
                                 'not found or already owned' %
                                 (job, self._workflow, self._instance))
                Reload._unown_tokens(client, result)
                return []
        return result

    def _own_all_job_tokens(self, client):
        assert not self._jobs
        name = Name(workflow=self._workflow, instance=self._instance)
        group_request = GroupRequest(namePrefix=name.get_job_prefix(),
                                     groupSuffix=Name.DELIMITER)
        group_response = client.group(group_request)
        counts = group_response.counts
        if not counts:
            self._output += ('workflow %s instance %s not found or already '
                             'archived' % (self._workflow, self._instance))
            return []
        assert len(counts) <= 2  # runnable and waiting
        num_job_tokens = 0
        for count in counts.values():
            num_job_tokens += count
        query = Query(namePrefix=name.get_job_prefix())
        query_and_own_request = QueryAndOwnRequest(
            owner='workflow_util',
            expirationTime=time.time() + Reload._LEASE_TIME_SEC,
            query=query)
        query_and_own_response = client.query_and_own(query_and_own_request)
        tokens = query_and_own_response.tokens
        assert len(tokens) <= num_job_tokens
        if len(tokens) < num_job_tokens:
            self._output += ('only %d out of %d job tokens in workflow %s '
                             'instance %s could be claimed' % (len(tokens),
                                                               num_job_tokens,
                                                               self._workflow,
                                                               self._instance))
            Reload._unown_tokens(client, tokens)
            return []
        return tokens

    def _update_job_tokens(self, job_tokens, config_parser):
        workflow_tokens = config_parser.get_workflow_tokens(self._workflow)
        new_job_tokens = {}
        for token in workflow_tokens:
            name = Name.from_job_token_name(token.name)
            if name.job:
                # Make sure it is a job not an event token.
                new_job_tokens[name.job] = token
        new_job_names = set(new_job_tokens.keys())
        job_names = set([Name.from_job_token_name(token.name).job for token in
                         job_tokens])
        missing_job_names = job_names.difference(new_job_names)
        if missing_job_names:
            self._output = 'jobs %s not found in workflow %s defined in %s' % (
                missing_job_names, self._workflow, str(PinballConfig.PARSER_PARAMS))
            return False
        for job_token in job_tokens:
            name = Name.from_job_token_name(job_token.name)
            assert name.job
            new_job_token = new_job_tokens[name.job]
            Reload._reload_job_token(job_token, new_job_token)
        return True

    def execute(self, client, store):
        config_parser = load_parser_with_caller(PinballConfig.PARSER,
                                                PinballConfig.PARSER_PARAMS,
                                                ParserCaller.WORKFLOW_UTIL)
        workflow_names = config_parser.get_workflow_names()
        if self._workflow not in workflow_names:
            return 'workflow %s not found in %s\n' % (
                self._workflow, str(PinballConfig.PARSER_PARAMS))
        if self._jobs:
            job_tokens = self._own_selected_job_tokens(client)
        else:
            job_tokens = self._own_all_job_tokens(client)
        if not job_tokens:
            return self._output
        if self._update_job_tokens(job_tokens, config_parser):
            self._output = ('reloaded %s in workflow %s instance %s' % (
                'jobs %s' % self._jobs if self._jobs else 'all jobs',
                self._workflow,
                self._instance))
        self._unown_tokens(client, job_tokens)
        return self._output


class Alter(Command):
    """Enable or disable jobs."""
    DISABLE, ENABLE = range(2)
    _MODE = None
    _MODE_NAMES = {
        DISABLE: 'disable',
        ENABLE: 'enable'
    }

    def __init__(self):
        self._workflow = None
        self._instance = None
        self._jobs = None
        self._force = None

    def prepare(self, options):
        self._force = options.force
        self._workflow = options.workflow
        self._instance = options.instance
        self._jobs = options.jobs.split(',')
        if not self._workflow or not self._instance or not self._jobs:
            raise CommandException('%s command takes name of workflow, '
                                   'instance, and a list of jobs' %
                                   Alter._MODE_NAMES[self._MODE])

    def _get_job_tokens(self, client):
        name = Name(workflow=self._workflow, instance=self._instance)
        prefix = name.get_job_prefix()
        query = Query(namePrefix=prefix)
        request = QueryRequest(queries=[query])
        response = client.query(request)
        # For a more efficient lookup.
        jobs = set(self._jobs)
        result = {}
        for job_token in response.tokens[0]:
            job_name = Name.from_job_token_name(job_token.name)
            if job_name.job in jobs:
                result[job_name.job] = job_token
        return result

    @staticmethod
    def _is_owned(token):
        return (token.owner and token.expirationTime and
                token.expirationTime > time.time() - 1)

    def _alter_job_tokens(self, job_tokens, client):
        request = ModifyRequest(updates=[])
        for job_token in job_tokens:
            # TODO(pawel): to prevent others from modifying the tokens while we
            # manipulate them, it would be better to own them.
            if Alter._is_owned(job_token):
                return False
            job = pickle.loads(job_token.data)
            if self._MODE == Alter.DISABLE:
                job.disabled = True
            elif self._MODE == Alter.ENABLE:
                job.disabled = False
            else:
                assert False, 'unrecognized mode %d' % self._MODE
            job_token.data = pickle.dumps(job)
            request.updates.append(job_token)
        try:
            client.modify(request)
        except:
            # This can happen because someone else could have updated one of
            # the tokens in the meantime.
            return False
        return True

    def execute(self, client, store):
        job_tokens = self._get_job_tokens(client)
        assert len(job_tokens) <= len(self._jobs)
        jobs_set = set(self._jobs)
        job_tokens_set = set(job_tokens.keys())
        output = ''
        delta = jobs_set.difference(job_tokens_set)
        mode = Alter._MODE_NAMES[self._MODE]
        if delta:
            return 'job(s) %s not found in the master.  Note that only jobs ' \
                   'of a running workflow can be %sd' % (list(delta), mode)
        message = '%s %d jobs in workflow %s instance %s' % (mode,
                                                             len(self._jobs),
                                                             self._workflow,
                                                             self._instance)
        output = ''
        if self._force or confirm(message):
            MAX_TRIES = 10
            i = 0
            while i < MAX_TRIES:
                if self._alter_job_tokens(job_tokens.values(), client):
                    break
                job_tokens = self._get_job_tokens(client)
                i += 1
            if i < MAX_TRIES:
                output += '%sd %d job(s) in %d tries\n' % (mode,
                                                           len(job_tokens),
                                                           i + 1)
            else:
                output += 'failed to %s job(s) in %d tries' % (mode, MAX_TRIES)
        return output


class Disable(Alter):
    _MODE = Alter.DISABLE


class Enable(Alter):
    _MODE = Alter.ENABLE


class Cleanup(Command):
    """Remove old workflows from the database."""
    def __init__(self):
        self._timestamp = None
        self._force = None

    def prepare(self, options):
        self._force = options.force
        if options.age_days < 7:
            raise CommandException('age of instances to remove must be at '
                                   'least 7 days')
        delta = datetime.timedelta(days=options.age_days).total_seconds()
        self._timestamp = time.time() - delta

    def execute(self, client, store):
        data_builder = DataBuilder(store, use_cache=True)
        workflows = data_builder.get_workflows()
        instances_to_delete = []
        directories_to_delete = []
        for workflow in workflows:
            instances = data_builder.get_instances(workflow.workflow)
            for instance in instances:
                if ((instance.status == Status.FAILURE or
                     instance.status == Status.SUCCESS) and instance.end_time
                        and instance.end_time < self._timestamp):
                    instances_to_delete.append(instance)
                    directories_to_delete.append(
                        get_logs_dir(instance.workflow, instance.instance))
                    if PinballConfig.S3_LOGS_DIR:
                        directories_to_delete.append(get_logs_dir(instance.workflow,
                                                                  instance.instance,
                                                                  PinballConfig.S3_LOGS_DIR))
        tokens_to_delete = []
        for instance in instances_to_delete:
            name = Name(workflow=instance.workflow, instance=instance.instance)
            tokens_to_delete += store.read_archived_tokens(
                name_prefix=name.get_instance_prefix())
        deleted_tokens = 0
        deleted_directories = 0
        output = ''
        if not tokens_to_delete:
            output += 'no tokens need to be cleaned up\n'
        else:
            print 'removing tokens:'
            for token in tokens_to_delete:
                print '\t%s' % token.name
            print 'removing directories:'
            for directory in directories_to_delete:
                print '\t%s' % directory
            message = 'remove %d tokens and %d directories' % (
                len(tokens_to_delete), len(directories_to_delete))
            if self._force or confirm(message):
                store.delete_archived_tokens(tokens_to_delete)
                deleted_tokens = len(tokens_to_delete)
                for directory in directories_to_delete:
                    self._delete_directory(directory)
                deleted_directories = len(directories_to_delete)
        output += 'removed %d token(s) and %d directory(ies)\n' % (
            deleted_tokens, deleted_directories)
        return output

    def _delete_directory(self, directory):
        """Delete the given directory."""
        if directory.startswith('s3n://'):
            delete_s3_directory(directory)
        else:
            shutil.rmtree(directory, ignore_errors=True)


class RebuildCache(Command):
    """Rebuild the data cache."""
    def __init__(self):
        self._force = None

    def prepare(self, options):
        self._force = options.force

    def execute(self, client, store):
        cache_size = len(store.read_cached_data_names())
        message = 'rebuild cache with %d data items' % cache_size
        if self._force or confirm(message):
            store.clear_cached_data()
            data_builder = DataBuilder(store, use_cache=True)
            data_builder.get_workflows()
            cache_size = len(store.read_cached_data_names())
            return 'rebuilt data cache.  It now has %d data items' % cache_size
        return ''


def _check_workflow_instances(store, workflow):
    """Check the number of running instances of the workflow.

    Args:
        store: The store to retrieve runs status.
        workflow: Name of the workflow.

    Returns:
        False if running instance number exceeds the max_running_instances setting;
        Otherwise, True.
    """
    data_builder = DataBuilder(store, use_cache=True)
    instances = data_builder.get_instances(workflow)
    schedule = data_builder.get_schedule(workflow)
    number_running_instances = sum(1 for instance in instances
                                   if instance.status == Status.RUNNING)
    if schedule.max_running_instances and number_running_instances \
            >= schedule.max_running_instances:
        return False
    return True


def run_command(options):
    command = _COMMANDS[options.command]()
    command.prepare(options)
    factory = Factory(master_hostname=PinballConfig.MASTER_HOST,
                      master_port=PinballConfig.MASTER_PORT)
    client = factory.get_client()

    # The reason why these imports are not at the top level is that some of the
    # imported code (db models initializing table names) depends on parameters
    # passed on the command line (master name).  Those imports need to be delayed
    # until after command line parameter parsing.
    from pinball.persistence.store import DbStore
    store = DbStore()
    return command.execute(client, store)


_COMMANDS = {'start': Start, 'stop': Stop, 'pause': Pause, 'resume': Resume,
             'retry': Retry, 'redo': Redo, 'poison': Poison, 'drain': Drain,
             'undrain': UnDrain, 'abort': Abort, 'unabort': UnAbort,
             'exit': Exit, 'unexit': UnExit, 'reschedule': ReSchedule,
             'unschedule': UnSchedule, 'reload': Reload, 'disable': Disable,
             'enable': Enable, 'cleanup': Cleanup,
             'rebuild_cache': RebuildCache}


def main():
    parser = argparse.ArgumentParser(
        description='Manipulate Pinball workflows.')
    parser.add_argument('-c',
                        '--config_file',
                        dest='config_file',
                        required=True,
                        help='full path to the pinball setting configure file')
    parser.add_argument('-f',
                        '--force',
                        dest='force',
                        action='store_true',
                        default=False,
                        help='do not ask for confirmation')
    parser.add_argument('-w',
                        '--workflow',
                        dest='workflow',
                        help='workflow name')
    parser.add_argument('-i',
                        '--instance',
                        dest='instance',
                        help='workflow instance')
    parser.add_argument('-j',
                        '--jobs',
                        dest='jobs',
                        help='comma separated list of jobs')
    parser.add_argument('-e',
                        '--execution',
                        dest='execution',
                        help='job execution number, 0-based, increasing from '
                             'the oldest to the most recent')
    parser.add_argument('-a',
                        '--age_days',
                        dest='age_days',
                        type=int,
                        default=28,
                        help='minimum age of workflow instances to clean up.')
    parser.add_argument('command',
                        choices=_COMMANDS.keys(),
                        help='command name')
    options = parser.parse_args(sys.argv[1:])

    PinballConfig.parse(options.config_file)

    if hasattr(PinballConfig, 'MASTER_NAME') and PinballConfig.MASTER_NAME:
        master_name(PinballConfig.MASTER_NAME)

    print run_command(options)


if __name__ == '__main__':
    main()



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Definition of a token used to generate unique version values.

Blessed version is stored in the master as any other token.  Each time a new
version number is needed, it is generated off the value stored in that token.
The value stored in blessed version is a monotonically increasing counter
so it is guaranteed that no single value is issued more than once.
"""

import sys
import time

from pinball.config.utils import timestamp_to_str
from pinball.master.thrift_lib.ttypes import Token


__author__ = 'Pawel Garbacki'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class BlessedVersion(Token):
    """A singleton token keeping track of token versions.

    Versions of tokens stored in a given master are required to be unique.
    """

    def __init__(self, name=None, owner=None):
        """Create blessed version with a given name and owner.

        Name and owner have to either both be set or none should be set.
        Blessed version in use should always have name and owner set.  The
        version of init with name and owner set to None relies on external
        initialization of those fields.
        Args:
            name: The name of the blessed version token.
            owner: The owner of the blessed version token.
        """
        assert (name and owner) or (not name and not owner)
        if name and owner:
            now = BlessedVersion._get_timestamp_millis()
            data_str = ('blessed version created at %s' %
                        timestamp_to_str(now / 1000))
            Token.__init__(self, now, name, owner, sys.maxint, 0, data_str)
        else:
            Token.__init__(self)

    @staticmethod
    def from_token(token):
        blessed_version = BlessedVersion()
        for key, value in token.__dict__.items():
            blessed_version.__dict__[key] = value
        return blessed_version

    @staticmethod
    def _get_timestamp_millis():
        """Return time in milliseconds since the epoch."""
        return int(time.time() * 1000)

    def advance_version(self):
        """Increase the internal version counter.

        The counter value is based on the current time.  Since those values
        are used as token modification ids, basing them on time has an
        advantage for debugging - looking at the version we can tell when a
        token was modified.

        A BIG WARNING: as an application developer do not assume anything about
        the semantics of version values other than their uniqueness.  The
        implementation details are subject to change.
        """
        self.version = max(self.version + 1,
                           BlessedVersion._get_timestamp_millis())
        return self.version



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Logic handling job execution.

In a typical case, each job class has its own executor.
"""
import abc
import atexit
import os
import select
import signal
import subprocess
import threading
import time
import traceback

from pinball.config.pinball_config import PinballConfig
from pinball.config.utils import get_log
from pinball.config.utils import timestamp_to_str
from pinball.persistence.token_data import TokenData
from pinball.workflow import log_saver
from pinball.workflow.job import ShellConditionJob
from pinball.workflow.job import ShellJob
from pinball.workflow.buffered_line_reader import BufferedLineReader
from pinball.workflow.utils import get_logs_dir


__author__ = 'Pawel Garbacki, Mao Ye'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


LOG = get_log('pinball.workflow.job_executor')


class ExecutionRecord(TokenData):
    """A data object holding information about a single job execution."""
    def __init__(self, info=None, instance=None, start_time=None,
                 end_time=None, exit_code=None, logs=None):
        self.info = info
        self.instance = instance
        self.start_time = start_time
        self.end_time = end_time
        self.exit_code = exit_code
        self.events = []
        # TODO(pawel): rename this to attributes for naming consistency.
        self.properties = {}
        self.cleanup_exit_code = None
        # Setting logs={} in the argument list is a bad idea.
        # See http://effbot.org/zone/default-values.htm for explanation.
        self.logs = logs if logs is not None else {}

    @property
    def _COMPATIBILITY_ATTRIBUTES(self):
        result = super(ExecutionRecord, self)._COMPATIBILITY_ATTRIBUTES
        result['properties'] = {}
        result['instance'] = None
        result['cleanup_exit_code'] = None
        result['events'] = []
        return result

    def get_event_attributes(self):
        ATTRIBUTE_PREFIX = 'EVENT_ATTR:'
        result = {}
        for key, value in self.properties.items():
            if key.startswith(ATTRIBUTE_PREFIX):
                result[key.split(':', 1)[1]] = value
        return result

    def __str__(self):
        if self.start_time:
            start_time = timestamp_to_str(self.start_time)
        else:
            start_time = str(self.start_time)
        if self.end_time:
            end_time = timestamp_to_str(self.end_time)
        else:
            end_time = str(self.end_time)
        return ('ExecutionRecord(info=%s, instance=%s, start_time=%s, '
                'end_time=%s, exit_code=%s, events=%s, properties=%s, '
                'logs=%s)' %
                (self.info, self.instance, start_time, end_time,
                 self.exit_code, self.events, self.properties, self.logs))

    def __repr__(self):
        return self.__str__()


class JobExecutor(object):
    """Interface of a client communicating with token master."""
    __metaclass__ = abc.ABCMeta

    def __init__(self, workflow, instance, job_name, job, data_builder,
                 emailer):
        self._workflow = workflow
        self._instance = instance
        self._job_name = job_name
        self.job = job
        self._data_builder = data_builder
        self._emailer = emailer
        # A map from log type to the log saver storing job output of this type.
        self._log_savers = {}

    _cleaners = set()

    @staticmethod
    @atexit.register
    def _call_cleaners():
        for cleaner in JobExecutor._cleaners:
            cleaner()

    @staticmethod
    def from_job(workflow, instance, job_name, job, data_builder, emailer):
        """Create an executor capable of running a given job."""
        if type(job) == ShellJob or type(job) == ShellConditionJob:
            return ShellJobExecutor(workflow, instance, job_name, job,
                                    data_builder, emailer)

    @abc.abstractmethod
    def prepare(self):
        return

    @abc.abstractmethod
    def execute(self):
        return

    @abc.abstractmethod
    def abort(self):
        return


class ShellJobExecutor(JobExecutor):
    def __init__(self, workflow, instance, job_name, job, data_builder,
                 emailer):
        super(ShellJobExecutor, self).__init__(workflow, instance, job_name,
                                               job, data_builder, emailer)
        # Indicates if job data has not been recorded in the master.
        self.job_dirty = False
        self._process = None
        self._aborted = False
        self._abort_timeout_reached = False
        self._warn_timeout_reached = False
        self._lock = threading.Lock()
        self._log_pipe_readers = {}

    def _get_logs_dir(self, log_directory):
        """Generate name of directory where job logs are stored.

        Returns:
            Name of the job logs directory.
        """
        return get_logs_dir(self._workflow, self._instance, log_directory)

    def _get_log_filename(self, log_type, timestamp):
        """Generate log file name.

        Args:
            log_type: Type of the log stored in the file.
            timestamp: The execution timestamp.
        """
        # TODO(pawel): the file name should contain more context, in particular
        # the workflow and instance names.
        filename = '%s.%d.%s' % (self.job.name, timestamp, log_type)
        log_directory = PinballConfig.S3_LOGS_DIR \
            if PinballConfig.S3_LOGS_DIR else PinballConfig.LOCAL_LOGS_DIR
        return os.path.join(self._get_logs_dir(log_directory), filename)

    def _get_last_execution_record(self):
        """Retrieve the most recent job execution record.

        Returns:
            The most recent job execution record.
        """
        assert self.job.history
        return self.job.history[-1]

    def _check_token_lost(self):
        """Check if the ownership of the job token has been lost.

        Returns:
            True iff the job token ownership has been lost.
        """
        if not self.job.history:
            return False
        execution_record = self.job.history[-1]
        assert execution_record.start_time
        if not execution_record.end_time:
            execution_record.end_time = time.time()
            execution_record.exit_code = 1
            message = 'executor failed to renew job ownership on time\n'
            self._append_to_pinlog(message)
            return True
        return False

    def prepare(self):
        """Prepare the execution.

        As a side effect, this method appends an execution record to the job
        history.

        Returns:
            True iff the preparations succeeded.  If False, the job should not
            be executed.
        """
        if self._check_token_lost():
            return False
        execution_record = ExecutionRecord(instance=self._instance,
                                           start_time=time.time())
        self.job.history.append(execution_record)
        self.job.truncate_history()
        execution_record.events = self.job.events
        self.job.events = []
        if self.job.disabled:
            execution_record.info = 'DISABLED'
            return True
        else:
            execution_record.info = self.job.command
        try:
            logs_dir = self._get_logs_dir(PinballConfig.LOCAL_LOGS_DIR)
            if not os.path.exists(logs_dir):
                os.makedirs(logs_dir)
        except:
            LOG.exception('')
            execution_record.end_time = time.time()
            execution_record.exit_code = 1
            return False

        for log_type in ['stdout', 'stderr']:
            execution_record.logs[log_type] = self._get_log_filename(
                log_type, execution_record.start_time)
            self._log_savers[log_type] = \
                log_saver.FileLogSaver.from_path(
                    execution_record.logs[log_type])
            self._log_savers[log_type].open()
        return True

    def _set_log_pipe_reader(self, process):
        """ Sets up buffered line reader to read the log pipes

        Args:
            process: the process to set up the log piper reader for
        """
        self._log_pipe_readers = {
            process.stdout: BufferedLineReader(process.stdout),
            process.stderr: BufferedLineReader(process.stderr)}

    def _process_log_line(self, line):
        """Process a log line extracting properties.

        Args:
            line: The log line to process.
        """
        # A magic value marking log lines with key=value pairs.
        PINBALL_MAGIC = 'PINBALL:'
        if line.startswith(PINBALL_MAGIC):
            if not line.endswith('\n'):
                LOG.error('PINBALL magic string is not properly terminated:')
                LOG.error(line)

            line = line.strip()
            line = line[len(PINBALL_MAGIC):]
            delimiter_index = line.find('=')
            key, value = (line[:delimiter_index], line[delimiter_index + 1:])
            if key:
                execution_record = self._get_last_execution_record()
                # Python guarantees that this operation is atomic so we don't
                # need locks around it.
                execution_record.properties[key] = value
                self.job_dirty = True

    def _consume_logs(self, process):
        """Process logs produced by the specified process.

        Args:
            fout: The file descriptor to write stdout to.
            ferr: The file descriptor to write stderr to.
            process: The process whose logs we want to consume.
        Returns:
            True iff any data was read.
        """
        TIMEOUT_SEC = 60.  # 1 minute
        streams = []
        if not process.stdout.closed:
            streams.append(process.stdout)
        if not process.stderr.closed:
            streams.append(process.stderr)
        if not streams:
            return False
        ready_to_read = select.select(streams,
                                      [],
                                      [],
                                      TIMEOUT_SEC)[0]
        if not ready_to_read:
            LOG.info('select timeout reached while reading output of command '
                     '%s', self.job.command)
        for source in ready_to_read:
            lines = self._log_pipe_readers[source].readlines()

            if self._log_pipe_readers[source].eof():
                source.close()

            for line in lines:
                if source == process.stdout:
                    self._log_savers['stdout'].write(line)
                else:
                    assert source == process.stderr
                    self._log_savers['stderr'].write(line)
                self._process_log_line(line)

        return ready_to_read != []

    def _write_separator_to_logs(self, flag):
        msg = '\n<<<<<<<<<<%s of cleanup code logs>>>>>>>>>>\n' % flag
        for log_type in ['stdout', 'stderr']:
            if log_type in self._log_savers:
                self._log_savers[log_type].write(msg)

    def _execute_cleanup(self):
        execution_record = self._get_last_execution_record()
        if not self.job.cleanup_template:
            return None
        try:
            cleanup_command = (self.job.cleanup_template %
                               execution_record.properties)
        except KeyError:
            message = ('Could not customize cleanup command %s with '
                       'properties %s' % (self.job.cleanup_template,
                                          execution_record.properties))
            self._append_to_pinlog(message)
            return 1
        env = os.environ.copy()
        env.pop('DJANGO_SETTINGS_MODULE', None)
        cleanup_process = subprocess.Popen(
            cleanup_command,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=True,
            env=env,
            preexec_fn=os.setsid)
        self._set_log_pipe_reader(cleanup_process)

        self._write_separator_to_logs('Start')
        while cleanup_process.poll() is None:
            self._consume_logs(cleanup_process)
        # Check again to catch anything after the process
        # exits.
        while self._consume_logs(cleanup_process):
            pass
        self._write_separator_to_logs('End')
        return cleanup_process.wait()

    def _append_to_pinlog(self, message):
        """Append message to pinlog.

        PINLOG means pinball log which contains log lines pinball itself
        generates for a particular job.

        Args:
            message: The log message to append.
        """
        execution_record = self._get_last_execution_record()
        pinlog = execution_record.logs.get('pinlog')
        try:
            if not pinlog:
                #TODO(Mao): Move the logic to create local dirs to log saver
                logs_dir = self._get_logs_dir(PinballConfig.LOCAL_LOGS_DIR)
                if not os.path.exists(logs_dir):
                    os.makedirs(logs_dir)
                pinlog = self._get_log_filename('pinlog',
                                                execution_record.start_time)
                self._log_savers['pinlog'] = log_saver.FileLogSaver.from_path(
                    pinlog)
                self._log_savers['pinlog'].open()
                execution_record.logs['pinlog'] = pinlog

            self._log_savers['pinlog'].write(message)
        except:
            LOG.exception('')

    def _get_emails(self):
        """Get notification emails for the currently running job.

        Returns:
            List of job's notification email addresses.
        """
        emails = set(self.job.emails)
        schedule_data = self._data_builder.get_schedule(self._workflow)
        if schedule_data:
            emails.update(schedule_data.emails)
        return list(emails)

    def _check_timeouts(self):
        """Check if timeouts have been reached."""
        if self._abort_timeout_reached:
            return
        execution_record = self._get_last_execution_record()
        start_time = execution_record.start_time
        now = time.time()
        if (self.job.abort_timeout_sec and
                start_time + self.job.abort_timeout_sec < now):
            self._abort_timeout_reached = True
            self.abort()
            return
        if (not self._warn_timeout_reached and self.job.warn_timeout_sec and
                start_time + self.job.warn_timeout_sec < now):
            self._warn_timeout_reached = True
            emails = self._get_emails()
            if not emails:
                return
            execution = len(self.job.history) - 1
            job_execution_data = self._data_builder.get_execution(
                self._workflow, self._instance, self._job_name, execution)
            self._emailer.send_job_timeout_warning_message(emails,
                                                           job_execution_data)

    def execute(self):
        """Execute the job.

        It is assumed that the prepare method has been called and returned
        True.

        Returns:
            True iff the execution succeeded.
        """
        execution_record = self._get_last_execution_record()
        assert not execution_record.end_time
        if self.job.disabled:
            execution_record.end_time = execution_record.start_time
            execution_record.exit_code = 0
            return True
        try:
            assert not self._process
            with self._lock:
                # We need the lock to prevent a situation where the job
                # executes even though it got aborted (by a different
                # thread).
                aborted = self._aborted
                if not aborted:
                    command = self.job.customize_command()
                    LOG.info('executing command: %s', command)
                    env = os.environ.copy()
                    # Pinball sets Django module path which may interfere
                    # with the command being executed.
                    env.pop('DJANGO_SETTINGS_MODULE', None)
                    # The os.setsid() is passed in the argument preexec_fn
                    # so it's run after the fork() and before  exec() to
                    # run the shell.  It attaches a session id of the child
                    # process to the parent process which is a shell in our
                    # case.  This will make it the group leader.  So when a
                    # signal is sent to the process group leader, it's
                    # transmitted to all of the child processes.
                    self._process = subprocess.Popen(
                        command,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                        shell=True,
                        env=env,
                        preexec_fn=os.setsid)
                    self._set_log_pipe_reader(self._process)
                    JobExecutor._cleaners.add(self.abort)

            if aborted:
                # TODO(pawel): we should have an explicit indicator that
                # the job was aborted.
                execution_record.exit_code = 1
            else:
                while self._process.poll() is None:
                    self._consume_logs(self._process)
                    self._check_timeouts()
                # Check again to catch anything after the process
                # exits.
                while self._consume_logs(self._process):
                    pass
                execution_record.exit_code = self._process.wait()
                JobExecutor._cleaners.remove(self.abort)
            with self._lock:
                self._process = None

            if execution_record.exit_code != 0:
                execution_record.cleanup_exit_code = \
                    self._execute_cleanup()
        except:
            LOG.exception('')
            execution_record.exit_code = 1
            self._append_to_pinlog(traceback.format_exc())
        finally:
            execution_record.end_time = time.time()
            # Make sure we've saved all the logs
            try:
                for log_type in self._log_savers:
                    self._log_savers[log_type].close()
            except:
                LOG.exception('')

        return execution_record.exit_code == 0

    def abort(self):
        """Abort the currently running job."""
        with self._lock:
            self._aborted = True
            if self._process:
                os.killpg(self._process.pid, signal.SIGKILL)



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities to construct and manipulate workflow token names.

Token names are hierarchical.  We use '/' as the level separator.

Job token is named
/workflow/<workflow_name>/<workflow_instance>/job/[waiting|runnable]/<job>

Event token is named
/workflow/<workflow_name>/<workflow_instance>/input/<job>/<input>/<event>
In a basic job dependency model, <input> is the name of an upstream job that
<job> depends on.  A special input indicating workflow start is defined for
jobs with no dependencies.

A signal token is named as follows (depending on the level of applicability):
 - top level: /workflow/__SIGNAL__/<action>
 - workflow level: /workflow/<workflow_name>/__SIGNAL__/<action>
 - workflow instance level:
      /workflow/<workflow_name>/<workflow_instance>/__SIGNAL__/<action>
"""
import re


__author__ = 'Pawel Garbacki'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class Name(object):
    def __init__(self, workflow=None, instance=None, job_state=None, job=None,
                 input_name=None, event=None, signal=None):
        """Create a name. """
        self.workflow = workflow
        self.instance = instance
        self.job_state = job_state
        self.job = job
        self.input = input_name
        self.event = event
        self.signal = signal

    DELIMITER = '/'

    WORKFLOW_PREFIX = '/workflow/'
    RUNNABLE_STATE = 'runnable'
    WAITING_STATE = 'waiting'
    # Special input defined for jobs with no dependencies.
    # TODO(pawel): we should think about this a bit more.  In particular, it
    # should be possible to have jobs with no dependencies which would re-run
    # as soon as they finish.
    # TODO(pawel): rather than having a special input name for top-level jobs,
    # we could simply make them runnable when instantiating the workflow.
    WORKFLOW_START_INPUT = '__WORKFLOW_START__'

    SCHEDULE_PREFIX = '/schedule/'
    WORKFLOW_SCHEDULE_PREFIX = '/schedule/workflow/'

    @staticmethod
    def from_workflow_prefix(prefix):
        result = Name()
        # TODO(pawel): for efficiency, regexes should be pre-compiled.
        TOKENS_REGEX = r'^/workflow/(?P<workflow>\w+)'
        m = re.match(TOKENS_REGEX, prefix)
        if m:
            result.workflow = m.group('workflow')
        return result

    @staticmethod
    def from_instance_prefix(prefix):
        result = Name()
        TOKENS_REGEX = r'^/workflow/(?P<workflow>\w+)/(?P<instance>\w+)'
        m = re.match(TOKENS_REGEX, prefix)
        if m:
            result.workflow = m.group('workflow')
            result.instance = m.group('instance')
        return result

    @staticmethod
    def from_job_prefix(prefix):
        result = Name()
        TOKENS_REGEX = (r'^/workflow/(?P<workflow>\w+)/(?P<instance>\w+)/job')
        m = re.match(TOKENS_REGEX, prefix)
        if m:
            result.workflow = m.group('workflow')
            result.instance = m.group('instance')
        return result

    @staticmethod
    def from_job_state_prefix(prefix):
        result = Name()
        TOKENS_REGEX = (r'^/workflow/(?P<workflow>\w+)/(?P<instance>\w+)/job/'
                        r'(?P<job_state>waiting|runnable)')
        m = re.match(TOKENS_REGEX, prefix)
        if m:
            result.workflow = m.group('workflow')
            result.instance = m.group('instance')
            result.job_state = m.group('job_state')
        return result

    @staticmethod
    def from_job_token_name(name):
        result = Name()
        TOKENS_REGEX = (r'^/workflow/(?P<workflow>\w+)/(?P<instance>\w+)/job/'
                        r'(?P<job_state>waiting|runnable)/(?P<job>\w+)$')
        m = re.match(TOKENS_REGEX, name)
        if m:
            result.workflow = m.group('workflow')
            result.instance = m.group('instance')
            result.job_state = m.group('job_state')
            result.job = m.group('job')
        return result

    @staticmethod
    def from_input_prefix(prefix):
        result = Name()
        TOKENS_REGEX = (r'^/workflow/(?P<workflow>\w+)/'
                        r'(?P<instance>\w+)/input/'
                        r'(?P<job>\w+)/(?P<input>\w+)')
        m = re.match(TOKENS_REGEX, prefix)
        if m:
            result.workflow = m.group('workflow')
            result.instance = m.group('instance')
            result.job = m.group('job')
            result.input = m.group('input')
        return result

    @staticmethod
    def from_event_token_name(name):
        result = Name()
        TOKENS_REGEX = (r'^/workflow/(?P<workflow>\w+)/'
                        r'(?P<instance>\w+)/input/'
                        r'(?P<job>\w+)/(?P<input>\w+)/(?P<event>\w+)$')
        m = re.match(TOKENS_REGEX, name)
        if m:
            result.workflow = m.group('workflow')
            result.instance = m.group('instance')
            result.job = m.group('job')
            result.input = m.group('input')
            result.event = m.group('event')
        return result

    @staticmethod
    def from_workflow_schedule_token_name(name):
        result = Name()
        TOKENS_REGEX = (r'^/schedule/workflow/(?P<workflow>\w+)$')
        m = re.match(TOKENS_REGEX, name)
        if m:
            result.workflow = m.group('workflow')
        return result

    @staticmethod
    def from_signal_token_name(name):
        result = Name()

        TOP_SIGNAL_REGEX = r'^/workflow/__SIGNAL__/(?P<signal>\w+)$'
        m = re.match(TOP_SIGNAL_REGEX, name)
        if m:
            result.signal = m.group('signal')
            return result

        WORKFLOW_SIGNAL_REGEX = (r'^/workflow/(?P<workflow>\w+)/__SIGNAL__/'
                                 r'(?P<signal>\w+)$')
        m = re.match(WORKFLOW_SIGNAL_REGEX, name)
        if m:
            result.workflow = m.group('workflow')
            result.signal = m.group('signal')
            return result

        INSTANCE_SIGNAL_REGEX = (r'^/workflow/(?P<workflow>\w+)/'
                                 r'(?P<instance>\w+)/__SIGNAL__/'
                                 r'(?P<signal>\w+)$')
        m = re.match(INSTANCE_SIGNAL_REGEX, name)
        if m:
            result.workflow = m.group('workflow')
            result.instance = m.group('instance')
            result.signal = m.group('signal')

        return result

    def get_workflow_prefix(self):
        if self.workflow:
            return '/workflow/%(workflow)s/' % {'workflow': self.workflow}
        return ''

    def get_instance_prefix(self):
        if self.workflow and self.instance:
            return ('/workflow/%(workflow)s/%(instance)s/' %
                    {'workflow': self.workflow, 'instance': self.instance})
        return ''

    def get_job_prefix(self):
        if self.workflow and self.instance:
            return ('/workflow/%(workflow)s/%(instance)s/job/' %
                    {'workflow': self.workflow, 'instance': self.instance})
        return ''

    def get_job_state_prefix(self):
        if self.workflow and self.instance and self.job_state:
            return ('/workflow/%(workflow)s/%(instance)s/job/%(job_state)s/' %
                    {'workflow': self.workflow,
                     'instance': self.instance,
                     'job_state': self.job_state})
        return ''

    def get_job_token_name(self):
        if (self.workflow and self.instance and self.job_state and
                self.job):
            return ('/workflow/%(workflow)s/%(instance)s/job/%(job_state)s/'
                    '%(job)s' % {'workflow': self.workflow,
                                 'instance': self.instance,
                                 'job_state': self.job_state,
                                 'job': self.job})
        return ''

    def get_input_prefix(self):
        if self.workflow and self.instance and self.job:
            return ('/workflow/%(workflow)s/%(instance)s/input/%(job)s/'
                    '%(input)s/' % {'workflow': self.workflow,
                                    'instance': self.instance,
                                    'job': self.job,
                                    'input': self.input})
        return ''

    def get_event_token_name(self):
        if (self.workflow and self.instance and self.job and
                self.input and self.event):
            return ('/workflow/%(workflow)s/%(instance)s/input/%(job)s/'
                    '%(input)s/%(event)s' % {'workflow': self.workflow,
                                             'instance': self.instance,
                                             'job': self.job,
                                             'input': self.input,
                                             'event': self.event})
        return ''

    def get_workflow_schedule_token_name(self):
        if self.workflow:
            return '/schedule/workflow/%(workflow)s' % {'workflow':
                                                        self.workflow}

    def get_signal_prefix(self):
        if not self.workflow:
            return '/workflow/__SIGNAL__/'
        if not self.instance:
            return '/workflow/%(workflow)s/__SIGNAL__/' % {
                'workflow': self.workflow}
        return ('/workflow/%(workflow)s/%(instance)s/__SIGNAL__/' % {
                'workflow': self.workflow,
                'instance': self.instance})

    def get_signal_token_name(self):
        if self.signal:
            if not self.workflow:
                return '/workflow/__SIGNAL__/%(signal)s' % {
                    'signal': self.signal}
            if not self.instance:
                return '/workflow/%(workflow)s/__SIGNAL__/%(signal)s' % {
                    'workflow': self.workflow,
                    'signal': self.signal}
            return ('/workflow/%(workflow)s/%(instance)s/__SIGNAL__/'
                    '%(signal)s' % {'workflow': self.workflow,
                                    'instance': self.instance,
                                    'signal': self.signal})
        return ''



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import time

from threading import Thread

from pinball.config.utils import get_log
from pinball.ui.data_builder import DataBuilder

LOG = get_log('pinball.ui.cache_thread')


__author__ = 'Julia Oh, Mao Ye'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = ['Julia Oh', 'Mao Ye']
__license__ = 'Apache'
__version__ = '2.0'


def start_cache_thread(dbstore):
    """Creates and starts a daemon thread for workflow data computation.

    This method is called when pinball ui server starts.

    Args:
        dbstore: The store to retrieve runs status.
    Returns:
        cache_thread
    """
    thread = Thread(target=_compute_workflow, args=[dbstore])
    thread.daemon = True
    thread.start()
    return thread


def _compute_workflow(dbstore):
    """Cache thread's target callable that computes the workflow.

    This runnable is called my thread's run() method when thread
    starts. It will compute workflows data, serialize it, and store it
    in _WORKFLOW_JSON. This computation will infinitely
    repeat itself, constantly updating the _WORKFLOW_JSON until pinball_ui
    server stops.

    Args:
        dbstore: The store to retrieve runs status.
    """
    global _WORKFLOWS_JSON
    data_builder = DataBuilder(dbstore, use_cache=True)
    while True:
        try:
            LOG.info("Workflow data computation starting.")
            workflows_data = data_builder.get_workflows()
            schedules_data = data_builder.get_schedules()
            _WORKFLOWS_JSON = _serialize(workflows_data, schedules_data)
            LOG.info("Workflow data computation complete.")
            # TODO(mao): Tune this parameter depending on future
            # pinball user experience.
            # TODO(mao): Make this computation run at scheduled time intervals
            # and cancel the next execution if the previous job hasn't
            # finished.
            time.sleep(60 * 20)
        except Exception as e:
            LOG.exception(e)


def _serialize(workflows_data, schedules_data):
    workflow_emails = {}
    workflows_info = []
    for schedule in schedules_data:
        workflow_emails[schedule.workflow] = schedule.emails
    for workflow in workflows_data:
        workflow_data = workflow.format()
        if workflow.workflow in workflow_emails:
            workflow_data['owners'] = ','.join(email.split('@')[0]
                                               for email in workflow_emails[workflow.workflow])
        else:
            workflow_data['owners'] = 'N/A'
        workflows_info.append(workflow_data)
    return json.dumps({'aaData': workflows_info})


def get_workflows_json():
    return _WORKFLOWS_JSON

_WORKFLOWS_JSON = _serialize([], [])




# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import pkgutil
import sys

from collections import defaultdict
from pinball_ext.common import utils


__author__ = 'Mao Ye, Changshu Liu'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


LOG = utils.get_logger('pinball_ext.common.import_utils')


class ModuleImport(object):
    """ModuleImport supports:
    1) import classes/modules for given import _import_directories
    2) retrieve class for given class_name
    3) retrieve all subclass names for a given parent class name
    """

    def __init__(self, import_directories,
                 base_class=object, map_class_module=False):
        if not import_directories:
            raise Exception("import directories are missing")
        self._import_directories = import_directories
        if base_class:
            self._base_class = base_class
        self._name_to_class_map = {}
        self._parent_child_class_name_map = defaultdict(list)
        self.name = self.__class__.__name__

        if map_class_module:
            self.class_name_to_module_name_map = {}
        else:
            self.class_name_to_module_name_map = None

    def get_class_by_name(self, class_name):
        """ Get class for a given name.

        Args:
            class_name: This is a string name of the class_name.

        Returns:
            The corresponding class.
        """
        return self._name_to_class_map.get(class_name, None)

    def get_subclass_name_list(self, base_class_name):
        """Get the subclass name list for a given base class name.

        Args:
            base_class_name: This is a string name of the base class.

        Returns:
            A list of subclass names.
        """
        return self._parent_child_class_name_map.get(base_class_name, None)

    def import_all_modules(self):
        """Import all modules for the given list of directories.

        Raises:
            ImportError: An error occurred when importing failed.
        """
        for dir_name in self._import_directories:
            try:
                self._import_modules_from_dir(dir_name, '')
            except ImportError:
                LOG.error('Error importing %s', dir_name)
                raise

    def _import_modules_from_dir(self, root_dir, file_suffix):
        """Import all modules from a given directory.

        Args:
            root_dir: This is directory that modules are imported from.
            file_suffix: This is a string representation of file suffix.

        Returns:
            There is no explicit return. Instead, the classes under the
            specified directory will be registered in _name_to_class_map.
        """

        def _import_all(module_dir):
            for importer, pkg_name, is_package in pkgutil.iter_modules([module_dir]):
                real_module_dir = module_dir.split('dist-packages/')[1] \
                    if 'dist-packages/' in module_dir else module_dir
                if is_package or pkg_name.endswith(file_suffix):
                    module_name = real_module_dir.replace('/', '.') + '.' + pkg_name
                    __import__(module_name)
                    self.register_module(module_name)
                if is_package:
                    _import_all(module_dir + '/' + pkg_name)
        # TODO(mao): To find a better way to handle import modules for dist-packages
        real_root_dir = root_dir.split('dist-packages/')[1] \
            if 'dist-packages/' in root_dir else root_dir
        root_module_name = real_root_dir.replace('/', '.')
        __import__(root_module_name)
        self.register_module(root_module_name)
        _import_all(root_dir)

    def register_module(self, module_name):
        """Register a particular module into _name_to_class_map and _parent_child_class_name_map.

           Args:
               module_name: string name of the module to be registered.

           Returns:
               There is no explict return.

           Raises:
               Exception: An exception occurred when the module's name is not
               unique.
        """

        for class_name, clazz in inspect.getmembers(sys.modules[module_name],
                                                    inspect.isclass):
            if not issubclass(clazz, self._base_class):
                continue
            existing_component = self._name_to_class_map.get(clazz.__name__)
            if existing_component and existing_component != clazz:
                raise Exception('Modules must have unique names. Duplicate: %s vs %s'
                                % (existing_component, clazz))

            self._name_to_class_map[clazz.__name__] = clazz

            if self.class_name_to_module_name_map is not None:
                self.class_name_to_module_name_map[clazz.__name__] = module_name

            # get the parent-child relationship
            parent_class_list = inspect.getmro(clazz)
            for p_class in parent_class_list:
                if p_class.__name__ is not clazz.__name__:
                    self._parent_child_class_name_map[p_class.__name__].append(clazz.__name__)

    def add_module(self, class_name, clazz):
        """Put class_name and corresponding class into _name_to_class_map.
        This is used in data_job.py.

           Args:
               class_name: This is a string name of the class.
               clazz: This is the class.

           Returns:
               There is no explicit return.
        """
        self._name_to_class_map[class_name] = clazz

    def module_exists(self, class_name):
        """Check if the give class_name is registered.

           Args:
               class_name: This is the string name of the class.

           Return:
               If the class is registered in _name_to_class_map, then True;
               otherwise, False.
        """
        return class_name in self._name_to_class_map

    def get_all_class_names(self):
        """Get a list of class names contained in _name_to_class_map."""
        return self._name_to_class_map.keys()

    def get_all_classes(self):
        """Get a list of classes contained in _name_to_class_map."""
        return self._name_to_class_map.values()



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base class for executors running Hadoop/Hive jobs."""

import datetime
import getpass
import os
import re
import subprocess

from pinball_ext.common import hadoop_utils
from pinball_ext.common.decorators import retry
from pinball_ext.common.utils import get_logger


__author__ = 'Zach Drach, Changshu Liu'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class ClusterExecutor(object):
    """ A class that implements Hadoop functionality across multiple
    Hadoop cluster types.

    Subclasses should implement the following functions on their platforms:
        - run_hive_query()
        - run_hadoop_job()
        - run_hadoop_streaming_job()
        - kill_job()
        - get_job_result()
        - get_job_resource_dir() [optional]

    The other public functions on the executor class make use of the
    above three functions, making it easy to implement a set of common
    Hadoop/Hive utility functions on multiple platforms.

    All Executor settings are stored in a Config object, described below.
    """
    class Config(object):
        """ An object for defining configuration parameters for this executor.

        All configuration parameters can be overridden in the constructor of the
        executor. These values are just the defaults.
        """
        # The user that is running the commands
        USER = getpass.getuser()

        # TODO(csliu): job name should be a job level property.
        # A label for all hadoop jobs started by the executor
        NAME = "AdHocCommand"

        # All jars under these dirs will be added as libjars. It could be local
        # path or s3/hdfs path, depending on how concrete executor interpret it.
        USER_LIBJAR_DIRS = []

        # This jar contains the main class of Hadoop app.
        USER_APPJAR_PATH = None

        # Archive path for Hadoop application.
        USER_ARCHIVE_PATH = None

        # Scheduling queue.
        SCHEDULER_QUEUE = 'prod_pool' if USER == 'prod' else None
        SCHEDULER_PARAM = None  # platform-specific

        # Name of the cluster platform.
        # pinball_ext.common.PLATFORM contains all platforms.
        PLATFORM = None

        # Configs about a node (usually the master node) inside a Hadoop cluster
        HADOOP_HOST_USER = hadoop_utils.HadoopHostConfig.USER_NAME
        HADOOP_HOST_NAME = hadoop_utils.HadoopHostConfig.HOST_NAME
        HADOOP_HOST_SSH_PORT = hadoop_utils.HadoopHostConfig.SSH_PORT
        HADOOP_HOST_SSH_KEY_FILE = hadoop_utils.HadoopHostConfig.SSH_KEY_FILE
        HADOOP_HOST_HOME = hadoop_utils.HadoopHostConfig.REMOTE_HADOOP_HOME

    def __init__(self, executor_config=None):
        """ Initialize parameters for the data job execution layer.

        Args:
            executor_config: a dictionary of configuration params that override
            the defaults. The key names should be the same as the field names of
            corresponding xxxExecutor.Config class.
        """
        executor_config = executor_config if executor_config else {}

        self.log = get_logger(self.__class__.__name__)

        # Override self.config using executor_config dict.
        self.config = self.Config()
        for key, value in executor_config.items():
            if key == 'USER_LIBJAR_DIRS':
                self.config.USER_LIBJAR_DIRS = value.split(',')
            else:
                setattr(self.config, key, value)

        # Construct HadoopHostConfig object according to overridden config.
        self.hadoop_host_config = self._contruct_hadoop_host_config()

        self.job_ids = []

    def _contruct_hadoop_host_config(self):
        hh_config = hadoop_utils.HadoopHostConfig()
        hh_config.USER_NAME = self.config.HADOOP_HOST_USER
        hh_config.HOST_NAME = self.config.HADOOP_HOST_NAME
        hh_config.SSH_PORT = self.config.HADOOP_HOST_SSH_PORT
        hh_config.SSH_KEY_FILE = self.config.HADOOP_HOST_SSH_KEY_FILE
        hh_config.REMOTE_HADOOP_HOME = self.config.HADOOP_HOST_HOME
        return hh_config

    @property
    def job_name(self):
        return "%s:%s" % (self.config.USER, self.config.NAME)

    def run_hive_query(self, query_str, upload_archive=True):
        """Run a hive query and return the raw results.

        Args:
            query_str: A hive query string.
            upload_archive: If true, we will upload the archive with git code
                before running this command. This is unnecessary for hive
                queries with no dependencies on our code base.

        Return:
            The tuple (output, stderr, job_ids)

            output: a list of rows of the query output. Each row is a
                list of strings.
            stderr: the job stderr in the same format
            job_ids: a list of hadoop job ids as python strings
        """
        raise NotImplementedError(
            "run_hive_query is not supported by this executor.")

    def run_hadoop_streaming_job(self,
                                 mapper,
                                 reducer,
                                 input_path,
                                 output_dir,
                                 partitioner=None,
                                 input_format='TextInputFormat',
                                 output_format='TextInputFormat',
                                 extra_args=None,
                                 extra_jars=None):
        """Run a Hadoop Streaming job on the cluster.

        See http://hadoop.apache.org/common/docs/r0.20.1/streaming.html.

        Args:
            mapper: The mapper command to execute.
            reducer: The reducer command to execute.
            input_path: The input file or directory on HDFS or S3.
            output_dir: Where to write the output on HDFS or S3.
            partitioner: Optional partitioner class name.
            input_format: The inputformat to use.
            output_format: The outputformat to use.
            extra_arguments: List of additional optional arguments to pass into
                the job. (see http://hadoop.apache.org/docs/mapreduce/r0.22.0/
                streaming.html#Specifying+Communication+Formats+in+Detail)
            extra_jars: List of additional jars to pass to hadoop streaming's
                libjars param.

        Returns:
            Same as run_hive_query

        Raises:
            CalledProcessError if job fails.
        """
        raise NotImplementedError(
            "run_hadoop_streaming_job is not supported by this executor.")

    def run_hadoop_job(self,
                       class_name,
                       jobconf_args=None,
                       extra_args=None,
                       extra_jars=None):
        """Runs a hadoop mapreduce job on the cluster.

        Returns stdout and stderr in list-of-lists format

        Args:
            class_name: Java class name for this Hadoop app. This must be in a
                jar defined in self.config.USER_APP_JAR.
            jobconf_args: a dictionary of -D<key>=<value> settings
            extra_args: list of additional args.
            extra_jars: List of extra jars to pass to Hadoop's libjars param.

        Returns:
            Same as run_hive_query

        Raises:
            CalledProcessError if the command fails.
        """
        raise NotImplementedError(
            "run_hadoop_job is not supported by this executor.")

    def kill_job(self, job_id):
        """Kills the Hadoop job with the given id."""
        raise NotImplementedError("kill_job is not supported by this executor.")

    def get_job_result(self, job_id):
        """Retrieves results for previously executed job."""
        raise NotImplementedError("get_job_result is not supported by this executor.")

    def get_job_resource_dir(self, run_as_user):
        """Path to location where job resources (jars, archives) are stored."""
        raise NotImplementedError("get_job_resource_dir is not supported by this executor.")

    def __str__(self):
        return self.__class__.__name__

    def __repr__(self):
        return str(self)

    @retry(subprocess.CalledProcessError, tries=3, delay=1, backoff=2)
    def get_table_description(self, table, database='default'):
        """Return table description string from Hive."""
        rows, stderr, job_ids = self.run_hive_query(
            "USE %s; DESCRIBE EXTENDED %s;" % (database, table),
            upload_archive=False)
        output = '\n'.join([' '.join(row) for row in rows])
        result = re.match(r'.*Detailed Table Information (.*)$',
                          output,
                          re.DOTALL)
        return result.groups()[0]

    @retry(subprocess.CalledProcessError, tries=3, delay=1, backoff=2)
    def get_table_location(self, table, database='default'):
        """Return the location of a table."""
        table_info = self.get_table_description(table, database=database)
        return re.match(r'.*location:([^,]*).*', table_info).groups()[0]

    @retry(subprocess.CalledProcessError, tries=3, delay=1, backoff=2)
    def get_table_locations(self, table, database='default', partition='dt'):
        """Return the locations for each partition of a table."""
        location = self.get_table_location(table, database=database)
        partitions = self.get_partitions(table, database=database)
        return [os.path.join(location, partition, '=', part) for part in partitions]

    @retry(subprocess.CalledProcessError, tries=3, delay=1, backoff=2)
    def drop_partition(self, table, database, partition_name, partition):
        """Drop a partition in given table in Hive."""
        self.run_hive_query(
            "USE %s; ALTER TABLE %s DROP PARTITION(%s='%s');" %
            (database, table, partition_name, partition),
            upload_archive=False)

    @retry(subprocess.CalledProcessError, tries=3, delay=1, backoff=2)
    def drop_partitions_condition(self, table, database, condition):
        """Drop partitions satisfying the given condition."""
        self.run_hive_query(
            "USE %s; ALTER TABLE %s DROP PARTITION(%s);" %
            (database, table, condition),
            upload_archive=False)

    @retry(subprocess.CalledProcessError, tries=3, delay=1, backoff=2)
    def recover_partitions(self, table, database='default'):
        """Recover partitions in given table in Hive."""
        self.run_hive_query(
            "USE %s; ALTER TABLE %s RECOVER PARTITIONS;" %
            (database, table),
            upload_archive=False)

    @retry(subprocess.CalledProcessError, tries=3, delay=1, backoff=2)
    def add_partition(self, table, database, partition_names, partition_values):
        """Add a partition in given table in Hive.

        Args:
            partition_names: an individual partition name or a list of partition
                names if the table has multiple partitions.
            partitions_values: the actual partition(s) corresponding to
                <partition_names>.

        Examples:
            add_partition(table, database, 'epoch', '2013-07-12-00-00')
            add_partition(table,
                          database,
                          ['epoch', 'action'], ['2013-07-12-00-00', 'a'])
        """
        if not isinstance(partition_names, (list, tuple)):
            partition_names = [partition_names]
        if not isinstance(partition_values, (list, tuple)):
            partition_values = [partition_values]
        if len(partition_names) != len(partition_values):
            raise ValueError('Unmatched partition param: %s vs %s' %
                             (partition_names, partition_values))

        partition_str = \
            ','.join(["%s='%s'" % t for t in zip(partition_names, partition_values)])
        self.run_hive_query(
            "USE %s; ALTER TABLE %s ADD IF NOT EXISTS PARTITION(%s);" %
            (database, table, partition_str),
            upload_archive=False)

    @retry(subprocess.CalledProcessError, tries=3, delay=1, backoff=2)
    def get_partitions(self, table, database='default'):
        """Return a list of epoch strings that Hive knows about for this table.

        This will return only the top partition for a multi-partitioned table.
        """
        rows, stderr, job_ids = self.run_hive_query(
            "USE %s; SHOW PARTITIONS %s;" % (database, table),
            upload_archive=False)
        return sorted(set([r[0].split('/')[0].split('=')[1] for r in rows]))

    @retry(subprocess.CalledProcessError, tries=3, delay=1, backoff=2)
    def get_full_partitions(self, table, database='default'):
        """Return a list of partitions for a table.

        Returns:
            a list of (<partition_names>, <partition_values>) tuples, the same
        format as accepted by the add_partition() function.
        """
        rows, stderr, job_ids = self.run_hive_query(
            "USE %s; SHOW PARTITIONS %s;" % (database, table),
            upload_archive=False)
        return [tuple(zip(*[pair.split('=') for pair in row[0].split('/')]))
                for row in rows]

    def get_available_dates(self, table, database='default'):
        """Return a list of datetime.dates that Hive knows for this table."""
        partitions = self.get_partitions(table, database=database)
        # Each row is of the form utc_date=<date>, so parse out the date part.
        dates = [datetime.datetime.strptime(r, "%Y-%m-%d").date()
                 for r in partitions if not r.startswith('_distcp')]
        return dates

    @retry(subprocess.CalledProcessError, tries=3, delay=1, backoff=2)
    def does_table_exist(self, table, database='default'):
        """Return True if the table exists in the database, else False."""
        rows, stderr, job_ids = self.run_hive_query(
            "USE %s; DESCRIBE EXTENDED %s;" % (database, table),
            upload_archive=False)
        for row in rows[0]:
            if row.startswith('Table %s does not exist' % table):
                return False
        return True

    def get_table_latest_date_partition(self, table, database='default'):
        """Get the latest date partition for the given table."""
        if not self.does_table_exist(table, database=database):
            return False
        existing_parts = self.get_available_dates(table, database=database)
        if not len(existing_parts):
            return None
        return existing_parts[-1]

    @retry(subprocess.CalledProcessError, tries=3, delay=1, backoff=2)
    def get_partition_description(self, table, partition, database='default'):
        """Return partition description string from Hive."""
        rows, stderr, job_ids = self.run_hive_query(
            "USE %s; DESCRIBE EXTENDED %s PARTITION( %s );" %
            (database, table, partition),
            upload_archive=False)
        return rows[-1][1]

    @retry(subprocess.CalledProcessError, tries=3, delay=1, backoff=2)
    def get_partition_location(self, table, partition, database='default'):
        """Return the location of table."""
        part_info = self.get_partition_description(table,
                                                   partition,
                                                   database=database)
        return re.match(r'.*location:([^,]*).*', part_info).groups()[0]

    @retry(subprocess.CalledProcessError, tries=3, delay=1, backoff=2)
    def alter_table_set_location(self, table, location, database='default'):
        """Set the table location in Hive.

        Args:
            table: table to set.
            location: path on S3 or HDFS.
        """
        self.run_hive_query("USE %s; ALTER TABLE %s SET LOCATION '%s';" %
                            (database, table, location),
                            upload_archive=False)

    @retry(subprocess.CalledProcessError, tries=3, delay=1, backoff=2)
    def alter_table_rename(self, table, new_name, database):
        """Rename the table in Hive.

        Args:
            table: table to rename.
            new_name: new name for the table.
        """
        self.run_hive_query("USE %s; ALTER TABLE %s RENAME TO %s;" %
                            (database, table, new_name),
                            upload_archive=False)

    ###########################################
    # Private helper functions
    ###########################################

    def _generate_hive_query_header(self, upload_archive=False):
        """Generates a string to set Hive environment variables.

        It requires that self._upload_archive() has been implemented by the
        subclass.

        Args:
            upload_archive: a boolean indicating whether the archive should be
                uploaded and added to the query header.
        """
        full_query_string = ''
        if upload_archive and self.config.USER_ARCHIVE_PATH:
            uploaded_archive_path = self._upload_archive()
            full_query_string += "add archive %s;\n" % uploaded_archive_path

        full_query_string += 'set mapred.job.name=%s;\n' % self.job_name

        return full_query_string

    def _upload_archive(self):
        """Uploads self.config.USER_ARCHIVE_PATH to the cluster and returns
        the path to the uploaded archive.

        The returned path may be an S3 path or a path on the Hadoop master
        node depending on the platform.

        file to upload: self.config.USER_ARCHIVE_PATH

        Returns:
            The path where self.config.USER_ARCHIVE_PATH is uploaded to.
        """
        raise NotImplementedError("Subclasses should implement this method")

    def _get_scheduler_job_setting(self):
        if not self.config.SCHEDULER_QUEUE:
            return ''
        else:
            return ' -D%s=%s ' % (self.config.SCHEDULER_PARAM,
                                  self.config.SCHEDULER_QUEUE)

    def _get_scheduler_hive_setting(self):
        if not self.config.SCHEDULER_QUEUE:
            return ''
        else:
            # Can't have a trailing space after the new line since the following
            # comment lines can't have a leading space.
            return ' SET %s=%s;\n' % (self.config.SCHEDULER_PARAM,
                                      self.config.SCHEDULER_QUEUE)

    def _get_scheduler_mrjob_setting(self):
        if not self.config.SCHEDULER_QUEUE:
            return ''
        else:
            return ' --jobconf="%s=%s" ' % (self.config.SCHEDULER_PARAM,
                                            self.config.SCHEDULER_QUEUE)



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pinball.persistence.store import Store


__author__ = 'Pawel Garbacki'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class EphemeralStore(Store):
    """Memory-only store used for testing."""
    def initialize(self):
        self._active_tokens = {}
        self._archived_tokens = {}
        self._cached_data = {}

    def commit_tokens(self, updates=None, deletes=None):
        updates = updates if updates is not None else []
        deletes = deletes if deletes is not None else []
        for token in updates:
            self._active_tokens[token.name] = token
        for token in deletes:
            del self._active_tokens[token.name]

    def delete_archived_tokens(self, deletes):
        for token in deletes:
            del self._archived_tokens[token.name]

    @staticmethod
    def _matches(name, name_prefix, name_infix, name_suffix):
        if name_prefix and not name.startswith(name_prefix):
            return False
        if name_infix and not name_infix in name:
            return False
        if name_suffix and not name.endswith(name_suffix):
            return False
        return True

    @staticmethod
    def _filter_tokens(tokens, name_prefix, name_infix, name_suffix):
        result = []
        for token in tokens:
            if EphemeralStore._matches(token.name, name_prefix, name_infix,
                                       name_suffix):
                result.append(token)
        return result

    def read_active_tokens(self, name_prefix='', name_infix='',
                           name_suffix=''):
        return EphemeralStore._filter_tokens(self._active_tokens.values(),
                                             name_prefix,
                                             name_infix,
                                             name_suffix)

    def read_archived_tokens(self, name_prefix='', name_infix='',
                             name_suffix=''):
        return EphemeralStore._filter_tokens(self._archived_tokens.values(),
                                             name_prefix,
                                             name_infix,
                                             name_suffix)

    def archive_tokens(self, tokens):
        for token in tokens:
            del self._active_tokens[token.name]
            self._archived_tokens[token.name] = token

    def get_cached_data(self, name):
        return self._cached_data.get(name)

    def set_cached_data(self, name, data):
        self._cached_data[name] = data

    def read_tokens(self, name_prefix='', name_infix='', name_suffix=''):
        return (EphemeralStore._filter_tokens(self._active_tokens.values(),
                                              name_prefix,
                                              name_infix,
                                              name_suffix) +
                EphemeralStore._filter_tokens(self._archived_tokens.values(),
                                              name_prefix,
                                              name_infix,
                                              name_suffix))

    def read_token_names(self, name_prefix='', name_infix='', name_suffix=''):
        tokens = self.read_tokens(name_prefix, name_infix, name_suffix)
        result = []
        for token in tokens:
            result.append(token.name)
        return result

    def read_archived_token_names(self, name_prefix='', name_infix='',
                                  name_suffix=''):
        tokens = self.read_archived_tokens(name_prefix, name_infix,
                                           name_suffix)
        result = []
        for token in tokens:
            result.append(token.name)
        return result

    def read_cached_data_names(self, name_prefix='', name_infix='',
                               name_suffix=''):
        cached_data_names = self._cached_data.keys()
        result = []
        for name in cached_data_names:
            if EphemeralStore._matches(name, name_prefix, name_infix,
                                       name_suffix):
                result.append(name)
        return result



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Logic handling log read/write."""

import abc
import os
import time

from pinball.common import s3_utils
from pinball.config.pinball_config import PinballConfig
from pinball.config.utils import get_log

LOG = get_log('pinball.workflow.log_saver')


__author__ = 'Mao Ye'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class LogSaver(object):
    """Interface of a component reading and writing job execution logs."""
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def write(self, content_str):
        """Write content_str to the file.

        Args:
            content_str: The string to be written to the file.
        """
        return

    @abc.abstractmethod
    def read(self):
        """Return the file content as string.

        Returns:
            content_str: The string of the file content.
        """
        return


class FileLogSaver(LogSaver):
    """FileLogSaver class provides basic methods interacting with a local file.

    A list of methods are open, write, read and close.

    Attributes:
        _file_path: A string of the local file path.
        _file_descriptor: A file descriptor of the file in _file_path.
    """
    def __init__(self, file_path):
        self._file_path = file_path
        self._file_descriptor = None

    def __str__(self):
        return self._file_path

    def open(self, mode='a+'):
        """Open the local file.

        Args:
            mode: The default mode is 'a+', i.e., append/read
        """
        self._file_descriptor = open(self._file_path, mode)

    def write(self, content_str):
        """Write to the local file.

        Args:
            content_str: The string to be written to the local file.
        """
        self._file_descriptor.write(content_str)
        self._file_descriptor.flush()

    def read(self):
        """Read content from the local file.

        Returns:
            String content of the local file.
        """
        return self._file_descriptor.read()

    def close(self):
        """Close the local file"""
        self._file_descriptor.close()
        self._file_descriptor = None

    @staticmethod
    def from_path(file_path):
        """A factory method which returns the right LogSaver class for the given file path.

        A list of supported file path includes:
        1) local file, which is supported by FileLogSaver;
        2) s3 file, which is supported by S3FileLogSaver.

        Args:
            file_path: A string presentation of the file path.

        Returns:
            A LogSaver class instance.
        """
        if file_path.startswith('s3n://'):
            return S3FileLogSaver(file_path)
        else:
            return FileLogSaver(file_path)


class S3FileLogSaver(FileLogSaver):
    """S3FileLogSaver class provides basic methods interacting with remote s3 file.

    A list of methods are open, close, read, write.
    Note that there is no appending operation in s3 file system. In order to implement
    write to s3, we write the content to local file first, then upload the content from
    local file to s3. Considering the write performance, we only upload the content from
    local file to s3, when either of the following two conditions is satisfied.

    Condition-1: it has been a long time since the last upload to s3
    Condition-2: there is a lot new content flushed into the local file, and waiting
    to be uploaded to s3

    Attributes:
        _S3_UPLOAD_INTERVAL_IN_SEC: A class level setting for Condition-1.
        _S3_UPLOAD_BATCH_IN_BYTE:  A class level setting for Condition-2.

        _file_path: The s3 file path to read from or write to.
        _s3_key: A Boto refernce to an s3 key for the _file_path
        _local_file_log_saver: This is a class instance of FileLogSaver,
        which helps to read and write content to a local file.

        _last_remote_upload_time: The last time (in second) when we uploaded the file to s3.
        _pending_bytes: The size of the data that is pending to be written to s3 file.
    """
    _S3_UPLOAD_BATCH_IN_BYTE = 1000
    _S3_UPLOAD_INTERVAL_IN_SEC = 3*60

    def __init__(self, file_path):
        super(S3FileLogSaver, self).__init__(file_path)
        local_file_path = self._file_path.replace(
            PinballConfig.S3_LOGS_DIR_PREFIX,
            PinballConfig.LOCAL_LOGS_DIR_PREFIX)
        self._local_file_log_saver = FileLogSaver(local_file_path)
        self._last_remote_upload_time = time.time()
        self._pending_bytes = 0L
        self._s3_key = None

    def open(self, mode=None):
        """Open S3FileLogSaver to make it ready to read/write.

        More specifically, we need a s3 key for ready/write to remote s3 file,
        and open the LogSaver for local file as well.

        """
        # TODO(Mao): With "a+" mode, we need to warn if local file is missing
        #  while there is a file in s3.
        self._s3_key = self._get_or_create_s3_key(self._file_path)

    def close(self):
        """Close S3FileLogSaver. No further operation on the saver are permitted.

        Note that, we need to make sure all the content which is stored in
        the local file is uploaded to s3.
        """
        self._sync_to_s3()
        self._s3_key = None

        LOG.info("deleting local file: %s as all content is uploaded.",
                 self._local_file_log_saver._file_path)
        if os.path.exists(self._local_file_log_saver._file_path):
            try:
                os.remove(self._local_file_log_saver._file_path)
            except OSError, e:
                LOG.warn('deletion failed due to: %s', e)

    def _check_s3_upload_condition(self):
        """Check whether to upload local log file to remote s3 storage.

        There are two conditions which are related to the class level attributes:
        _S3_UPLOAD_BATCH_IN_BYTE and _S3_UPLOAD_INTERVAL_IN_SEC.

        Returns:
            True: If either of the two conditions is satisfied.
        """
        if self._pending_bytes >= self._S3_UPLOAD_BATCH_IN_BYTE:
            return True
        elif time.time() - self._last_remote_upload_time >= self._S3_UPLOAD_INTERVAL_IN_SEC:
            return True

    def write(self, content_str):
        """Write the content_str to remote s3 storage.

        Since there is no appending operation in s3 storage,
        we write the content_str to local file, and then upload
        the local file to the remote s3 storage.

        Args:
            content_str: The string to be written to s3.
        """

        # First write the content_str to local file
        self._local_file_log_saver.open()
        self._write_to_local_file(content_str)
        self._local_file_log_saver.close()

        # Check if we need to upload the local file to remote s3 storage.
        self._pending_bytes += len(content_str)
        if self._check_s3_upload_condition():
            self._sync_to_s3()
            self._last_remote_upload_time = time.time()
            self._pending_bytes = 0L

    def read(self):
        """Read from a s3 file."""
        return self._s3_key.get_contents_as_string()

    def _sync_to_s3(self):
        """Upload data from local file to remote s3 storage."""
        self._local_file_log_saver.open()
        content = self._local_file_log_saver.read()
        self._local_file_log_saver.close()
        self._s3_key.set_contents_from_string(content)
        LOG.info("%d bytes of data has been uploaded to s3 path %s",
                 len(content),
                 self._file_path)

    def _write_to_local_file(self, content_str):
        """Write content_str to a local file."""
        self._local_file_log_saver.write(content_str)

    @staticmethod
    def _get_or_create_s3_key(s3_location):
        """Get or create a Boto reference to an s3 key of the s3_location."""
        bucket_name, path = s3_utils.parse_s3_location(s3_location)
        bucket = s3_utils.get_s3_bucket(bucket_name)
        key = bucket.get_key(path)
        if not key:
            key = bucket.new_key(path)
        return key




# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pinball_ext.job import basic_jobs
from pinball_ext.common import import_utils


__author__ = 'Changshu Liu'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


_JOB_MODULES = None


def get_job_modules(import_dirs):
    global _JOB_MODULES
    if not _JOB_MODULES:
        _JOB_MODULES = import_utils.ModuleImport(import_dirs,
                                                 basic_jobs.JobBase)
        _JOB_MODULES.import_all_modules()
    return _JOB_MODULES


def get_py_class_by_name(job_import_dirs, job_class_name):
    job_modules = get_job_modules(job_import_dirs)
    return job_modules.get_class_by_name(job_class_name)


def get_sorted_job_names(job_import_dirs):
    job_modules = get_job_modules(job_import_dirs)
    job_names = job_modules.get_all_class_names()
    return sorted(job_names)



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Validation tests for parser extracting configs from a repository."""
import mock
import pickle
import unittest

from pinball.master.thrift_lib.ttypes import Token
from pinball.parser.repository_config_parser import \
    RepositoryConfigParser
from pinball.repository.config import JobConfig
from pinball.repository.config import WorkflowScheduleConfig
from pinball.scheduler.overrun_policy import OverrunPolicy
from pinball.workflow.name import Name


__author__ = 'Pawel Garbacki'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class SomeJobTemplate(object):
    def __init__(self, name, max_attempts=None, emails=None, priority=None):
        assert name == 'some_job'
        assert max_attempts == 10
        assert emails == ['some_email@pinterest.com',
                          'some_other_email@pinterest.com']
        assert priority == 100

    def get_pinball_job(self, inputs, outputs, params=None):
        assert inputs == ['__WORKFLOW_START__']
        assert outputs == []
        assert params == {'some_param': 'some_value'}

        return Token(name='/workflow/some_workflow/job/waiting/some_job')


class RepositoryConfigParserTestCase(unittest.TestCase):
    def setUp(self):
        self._call_count = 0

    @mock.patch('time.time')
    @mock.patch('pinball.parser.repository_config_parser.'
                'GithubRepository')
    def test_get_schedule_token(self, repository_mock, time_mock):
        repository = mock.Mock()
        repository_mock.return_value = repository

        schedule_config = WorkflowScheduleConfig()
        schedule_config.workflow = 'some_workflow'
        schedule_config.start_date = '2012-01-01'
        schedule_config.time = '00.00.01.000'
        schedule_config.recurrence = '1d'
        schedule_config.overrun_policy = 'START_NEW'
        schedule_config.emails = ['some_email@pinterest.com',
                                  'some_other_email@pinterest.com']
        repository.get_schedule.return_value = schedule_config

        time_mock.return_value = 100.  # the value must be a float

        parser = RepositoryConfigParser()

        schedule_token = parser.get_schedule_token('some_workflow')

        self.assertEqual('/schedule/workflow/some_workflow',
                         schedule_token.name)
        # 1325376000 = 01 Jan 2012 00:00:00 UTC
        self.assertEqual(1325376000, schedule_token.expirationTime)
        schedule = pickle.loads(schedule_token.data)
        self.assertEqual(1325376000, schedule.next_run_time)
        self.assertEqual(24 * 60 * 60, schedule.recurrence_seconds)
        self.assertEqual(OverrunPolicy.START_NEW, schedule.overrun_policy)
        self.assertEqual('some_workflow', schedule.workflow)
        self.assertEqual(['some_email@pinterest.com',
                          'some_other_email@pinterest.com'], schedule.emails)

        repository.get_schedule.assert_called_once_with('some_workflow')

    @mock.patch('pinball.parser.repository_config_parser.'
                'GithubRepository')
    def test_get_workflow_tokens(self, repository_mock):
        repository = mock.Mock()
        repository_mock.return_value = repository

        repository.get_job_names.return_value = ['some_job']

        job_config = JobConfig()
        job_config.workflow = 'some_workflow'
        job_config.job = 'some_job'
        job_config.is_condition = False
        job_config.template = ('tests.pinball.parser.'
                               'repository_config_parser_test.SomeJobTemplate')
        job_config.template_params = {'some_param': 'some_value'}
        job_config.parents = []
        job_config.emails = ['some_email@pinterest.com',
                             'some_other_email@pinterest.com']
        job_config.max_attempts = 10
        job_config.retry_delay_sec = 20
        job_config.priority = 100
        repository.get_job.return_value = job_config

        parser = RepositoryConfigParser()

        workflow_tokens = parser.get_workflow_tokens('some_workflow')

        self.assertEqual(2, len(workflow_tokens))

        # Verify the triggering event token.
        if Name.from_event_token_name(workflow_tokens[0].name).workflow:
            event_token = workflow_tokens[0]
        else:
            event_token = workflow_tokens[1]

        event_name = Name.from_event_token_name(event_token.name)
        self.assertEqual('some_workflow', event_name.workflow)
        self.assertEqual('some_job', event_name.job)
        self.assertEqual('__WORKFLOW_START__', event_name.input)

        event = pickle.loads(event_token.data)
        self.assertEqual('repository_config_parser', event.creator)

        repository.get_job_names.assert_called_once_with('some_workflow')
        repository.get_job.assert_called_once_with('some_workflow', 'some_job')



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Snapshot maintains a collection of tokens matching a given query."""


__author__ = 'Pawel Garbacki'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class Snapshot(object):
    def __init__(self, client, request):
        self._client = client
        self._request = request
        self._response = None
        self.refresh()

    def refresh(self):
        """Query the master.

        Returns:
            True if the local copy of the tokens has changed.  Otherwise False.
        """
        old_response = self._response
        self._response = self._client.query(self._request)
        return self._response != old_response



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
import sys
import unittest

from pinball_ext.job import hive_jobs
from pinball_ext.job import job_runner


__author__ = 'Changshu Liu'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'

JOB_IMPORT_DIR = [
    'tests/pinball_ext/job'
]

_RUNNER_ARGS = [
    '--job_class_name', 'JobRunnerTestHiveJob',
    '--job_params', 'start_date=2015-01-01,end_date=2015-01-07',
    '--executor', 'emr',
    '--executor_config', 'USER_APPJAR_PATH=s3n:/my_bucket/my_dir/my_jar.jar',
    '--job_import_dirs_config', 'tests.pinball_ext.job.job_runner_test.JOB_IMPORT_DIR',
]

_Q_TEMPLATE = "SELECT * FROM unknown_db.unknown_tb WHERE dt >= '%(start_date)s' AND dt <= '%(end_date)s'"

_PARAMS = {
    'start_date': '2015-01-01',
    'end_date': '2015-01-07',
}


class JobRunnerTestHiveJob(hive_jobs.HiveJob):
    _QUERY_TEMPLATE = _Q_TEMPLATE


def _mock_emr_run_hive_query_job(q_str, upload_archive):
    print(q_str)
    return None, None, []


class JobRunnterTestCase(unittest.TestCase):
    @mock.patch('pinball_ext.job.basic_jobs.ClusterJob._is_prod_job',
                mock.Mock(return_value=False))
    @mock.patch('pinball_ext.executor.emr_executor.EMRExecutor.__init__',
                mock.Mock(return_value=None))
    @mock.patch('pinball_ext.executor.emr_executor.EMRExecutor.run_hive_query',
                side_effect=_mock_emr_run_hive_query_job)
    def test_job_runner_run_job(self, f1):
        sys.argv = [sys.argv[0]] + _RUNNER_ARGS
        job_runner.main()
        f1.assert_called_once_with(_Q_TEMPLATE % _PARAMS, upload_archive=False)



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generic utils shared across pinball modules."""

import base64
import calendar
import datetime
import logging
import os
import pickle
import pytz
import random
import re
import socket
import sys
import threading
import time

from django.conf import settings
from pinball.config.pinball_config import PinballConfig


__author__ = 'Pawel Garbacki'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class PinballException(Exception):
    def __init__(self, message):
        Exception.__init__(self, message)


def get_log(module):
    logging.basicConfig(
        format='[%(asctime)s] - %(name)s - %(levelname)-8s"%(message)s"',
        datefmt='%Y-%m-%d %a %H:%M:%S')
    log = logging.getLogger(module)
    log.setLevel(logging.INFO)
    return log


def sanitize_name(name):
    """Rewrite the name to contain only alphanumerics and underscores.
    This function is used mainly to sanitize token names to simplify their
    parsing.

    Args:
        name: The name to rewrite.
    Returns:
        A rewritten version of the name.  If the original name does not contain
        any forbidden characters, it is returned without change.
    """
    ALLOWED_FORMAT = r'^\w+$'
    if re.match(ALLOWED_FORMAT, name):
        return name
    return base64.b32encode(name).replace('=', '_')


def _check_name_sanitized(name):
    """Check if a name is sanitized.

    Args:
        name: The name to check.
    Raises:
        PinballException: If name is not sanitized.  The message in the
        exception explains why the check failed.
    """
    sanitized_name = sanitize_name(name)
    if sanitized_name != name:
        raise PinballException('name %s contains disallowed characters.  Name '
                               'may contain only alphanumerics and '
                               'underscores' % name)


# The global store for master name.
_MASTER_NAME = sanitize_name(socket.gethostname())


def master_name(new_master_name=None):
    """Get token master name.

    In a typical case, token master name does not change between restarts of a
    given master server.  This name is used by the persistence layer to
    identify the location of the master state.  This way master can recover
    from failures.

    Args:
        new_master_name: The master name to return in subsequent calls of this
            function.
    Returns:
        A master name.  It defaults to the name of the local host.
    Raises:
        PinballException: If new_master_name is not sanitized.
    """
    global _MASTER_NAME
    if new_master_name:
        _check_name_sanitized(new_master_name)
        _MASTER_NAME = new_master_name
    return _MASTER_NAME


def timestamp_to_str(timestamp):
    """Convert epoch timestamp to a time string in UTC.

    Args:
        timestamp: The epoch time, in seconds, to convert.
    Returns:
        String representation of the timestamp in UTC timezone.
    """
    if not timestamp:
        return ''
    if timestamp == sys.maxint:
        # sys.maxint represents infinity.
        return 'inf'
    utc = pytz.timezone('UTC')
    return datetime.datetime.fromtimestamp(timestamp, tz=utc).strftime(
        '%Y-%m-%d %H:%M:%S %Z')


def str_to_timestamp(time_str):
    """Convert a time string to epoch timestamp.

    Args:
        time_str: String representation of the timestamp in UTC timezone.
    Returns:
        The epoch time, in seconds.
    """
    date = datetime.datetime.strptime(time_str, '%Y-%m-%d %H:%M:%S %Z')
    return calendar.timegm(date.utctimetuple())


def get_unique_name():
    """Generate a globally unique name.

    Returns:
        A name that is unique across threads, processes, machines, and time.
        The name contains only alphanumerics and underscores.
    """
    # TODO(pawel): make sure this is truly unique e.g., by memorizing all names
    # issued during the last millisecond or whatever the resolution of
    # time.time() is.
    hostname = sanitize_name(socket.gethostname())
    thread_name = sanitize_name(threading.current_thread().name)
    return '%s_%d_%d_%s_%d_%d_%d' % (hostname,
                                     PinballConfig.GENERATION,
                                     os.getpid(),
                                     thread_name,
                                     threading.current_thread().ident,
                                     int(time.time() * 1000),
                                     int(random.random() * sys.maxint))


def token_data_to_str(token_data):
    try:
        return str(pickle.loads(token_data))
    except:
        return token_data


def token_to_str(token):
    data_str = token_data_to_str(token.data)
    if token.expirationTime:
        if token.expirationTime == sys.maxint:
            expiration = 'inf (%d)' % token.expirationTime
        else:
            expiration = timestamp_to_str(token.expirationTime)
    else:
        expiration = str(token.expirationTime)
    return ('Token(version=%d, owner=%s, expirationTime=%s, priority=%f, '
            'name=%s, data=%s)' % (token.version, token.owner, expiration,
                                   token.priority, token.name, data_str))


def set_django_environment():
    settings.configure(DEBUG=PinballConfig.DEBUG,
                       ALLOWED_HOSTS=PinballConfig.ALLOWED_HOSTS,
                       SECRET_KEY=PinballConfig.SECRET_KEY,
                       INSTALLED_APPS=PinballConfig.INSTALLED_APPS,
                       MIDDLEWARE_CLASSES=PinballConfig.MIDDLEWARE_CLASSES,
                       ROOT_URLCONF=PinballConfig.ROOT_URLCONF,
                       TEMPLATE_DIRS=PinballConfig.TEMPLATE_DIRS,
                       STATICFILES_DIRS=PinballConfig.STATICFILES_DIRS,
                       STATIC_ROOT=PinballConfig.STATIC_ROOT,
                       MANAGERS=PinballConfig.MANAGERS,
                       STATIC_URL=PinballConfig.STATIC_URL,
                       TEMPLATE_CONTEXT_PROCESSORS=
                       PinballConfig.TEMPLATE_CONTEXT_PROCESSORS,
                       DATABASES=PinballConfig.DATABASES)



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

import mock
from pinball_ext.common.hadoop_utils import HadoopHostConfig
from pinball_ext.executor import qubole_executor
from pinball_ext.executor.common import Platform


__author__ = 'Changshu Liu'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


hadoop_node = HadoopHostConfig()
hadoop_node.USER_NAME = 'test_user'
hadoop_node.HOST_NAME = 'test_host_name'
hadoop_node.SSH_PORT = 22
hadoop_node.SSH_KEY_FILE = '~/test.key'
hadoop_node.HOST_NAME = '/usr/hadoop/'


def _mock_run_qubole_cmd(qb_cmd, log_line, kwargs):
    return None, 'stdout', 'stderr', []


def _mock_list_s3_directory(s3_dir):
    print(s3_dir)
    if s3_dir[-1] == '1':
        return ['%s/jar10.jar' % s3_dir, '%s/jar11.jar' % s3_dir]
    elif s3_dir[-1] == '2':
        return ['%s/jar20.jar' % s3_dir, '%s/jar21.jar' % s3_dir]
    return []


def _mock_run_shell_command(qb_cmd):
    return None, 'stdout', 'stderr', []


class QuboleExecutorTest(unittest.TestCase):
    @mock.patch('pinball_ext.common.utils.get_random_string',
                mock.Mock(return_value='random_str'))
    @mock.patch('pinball_ext.common.s3_utils.list_s3_directory',
                side_effect=_mock_list_s3_directory)
    @mock.patch('pinball_ext.executor.qubole_executor.QuboleExecutor.run_shell_command',
                side_effect=_mock_run_shell_command)
    def test_hadoop_job(self, run_command, s3_list):
        user_jar_dirs = ['s3://b1/dir1', 's3://b2/dir2']
        user_app_jar = 's3:/b/dir/jar.jar'
        user_archive = 's3://b3/dir3/file.archive'

        executor = qubole_executor.QuboleExecutor(
            executor_config={
                'QUBOLE_JARS_BLACKLIST': 'hadoop.jar,hive.jar,jar10.jar',
                'USER_LIBJAR_DIRS': ','.join(user_jar_dirs),
                'USER_APPJAR_PATH': user_app_jar,
                'USER_ARCHIVE_PATH': user_archive,
                'USER': 'test_user',
            }
        )
        self.assertEqual(executor.config.PLATFORM, Platform.QUBOLE)

        executor.run_hadoop_job(
            'test_job_class',
            jobconf_args={
                'pinball.key1': 'value1',
                'pinball.key2': 'value2',
            },
            extra_args=[
                '-Dpinball.key3=value3'
            ])

        self.assertEqual(run_command.call_count, 1)
        self.assertEquals(len(run_command.call_args_list[0][0]), 1)
        self.assertEquals(run_command.call_args_list[0][0][0],
                          "mkdir -p /tmp/hadoop_users/test_user/random_str && "
                          "hadoop fs -get s3://b1/dir1 /tmp/hadoop_users/test_user/random_str && "
                          "hadoop fs -get s3://b2/dir2 /tmp/hadoop_users/test_user/random_str && "
                          "hadoop fs -get s3:/b/dir/jar.jar /tmp/hadoop_users/test_user/random_str/jar.jar && "
                          "rm -f /tmp/hadoop_users/test_user/random_str/hadoop.jar && "
                          "rm -f /tmp/hadoop_users/test_user/random_str/hive.jar && "
                          "rm -f /tmp/hadoop_users/test_user/random_str/jar10.jar && "
                          "export HADOOP_CLASSPATH=/tmp/hadoop_users/test_user/random_str/* && "
                          "hadoop jar /tmp/hadoop_users/test_user/random_str/jar.jar test_job_class "
                          "-libjars /tmp/hadoop_users/test_user/random_str/jar21.jar"
                          ",/tmp/hadoop_users/test_user/random_str/jar11.jar"
                          ",/tmp/hadoop_users/test_user/random_str/jar20.jar "
                          "-Dmapred.job.name=test_user:AdHocCommand "
                          "-Dpinball.key2=value2 "
                          "-Dpinball.key1=value1 "
                          "-Dpinball.key3=value3;\n"
                          "EXIT_CODE=$?; \n"
                          "rm -rf /tmp/hadoop_users/test_user/random_str; \n"
                          "exit $EXIT_CODE;")

    @mock.patch('pinball_ext.executor.qubole_executor.QuboleExecutor._run_qubole_command_with_stderr',
                side_effect=_mock_run_qubole_cmd)
    def test_hive_job(self, run_qubole_cmd):
        executor = qubole_executor.QuboleExecutor(
            executor_config={
                'USER': 'test_user',
            }
        )

        q_str = 'SHOW TABLES;'
        executor.run_hive_query(q_str)

        self.assertEquals(run_qubole_cmd.call_count, 1)
        self.assertEquals(len(run_qubole_cmd.call_args_list[0][0]), 3)
        self.assertEquals(run_qubole_cmd.call_args_list[0][0][2], {
            'query': 'set mapred.job.name=test_user:AdHocCommand;\nSHOW TABLES;'})



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Constants and functions to be shared among all the executors."""


__author__ = 'Changshu Liu'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class Platform(object):
    """Enum that defines the available platforms."""
    EMR = 'emr'
    LOCAL = 'local'
    QUBOLE = 'qubole'


def make_local_executor(executor_config):
    # NOTE: import here to avoid cyclic import
    from pinball_ext.executor import local_executor
    return local_executor.LocalExecutor(executor_config)


def make_emr_executor(executor_config):
    # NOTE: import here to avoid cyclic import
    from pinball_ext.executor import emr_executor
    return emr_executor.EMRExecutor(executor_config)


def make_qubole_executor(executor_config):
    # NOTE: import here to avoid cyclic import
    from pinball_ext.executor import qubole_executor
    return qubole_executor.QuboleExecutor(executor_config)


EXECUTOR_FACTORY_REGISTRY = {
    Platform.EMR: make_emr_executor,
    Platform.LOCAL: make_local_executor,
    Platform.QUBOLE: make_qubole_executor,
}


def make_executor(executor_name, executor_config):
    if executor_name not in EXECUTOR_FACTORY_REGISTRY:
        raise KeyError("Executor with name: %s does not exist, available "
                       "executors are: %s. " % (
                       executor_name,
                       ','.join(EXECUTOR_FACTORY_REGISTRY.keys())))
    executor_factory = EXECUTOR_FACTORY_REGISTRY[executor_name]
    executor_config = executor_config.copy()
    return executor_factory(executor_config)
